import logging

import tiktoken
import requests.exceptions
from ._env import env


class CantLoadTikTokenEncoding(RuntimeError):
    ...


def _resolve_tiktoken_encoding(
    for_model: str = None, encoding: str | tiktoken.Encoding = None
) -> tiktoken.Encoding:
    assert (
        for_model is None or encoding is None
    ), "You may specify encoding or for_model(LLM), but not both"
    if isinstance(encoding, tiktoken.Encoding):
        return encoding
    if for_model is None and encoding is None:
        if env().config.TIKTOKEN_ENCODING:
            return _resolve_tiktoken_encoding(encoding=env().config.TIKTOKEN_ENCODING)
        for_model = (
            env().config.LLM_DEFAULT_ARGS.get("model", None) or env().config.MODEL
        )
    if for_model:
        try:
            if for_model.startswith("gpt-4.1") or for_model.startswith("gpt-4.5"):
                return tiktoken.get_encoding("o200k_base")
            return tiktoken.encoding_for_model(for_model)
        except (KeyError, requests.exceptions.ConnectionError):
            logging.warning(
                f"Can't resolve tiktoken encoding for '{for_model}'. "
                f"Default encoding will be used."
            )
    encoding = encoding or "cl100k_base"
    try:
        return tiktoken.get_encoding(encoding)
    except (ValueError, requests.exceptions.ConnectionError) as e:
        raise CantLoadTikTokenEncoding(
            f"Can't load tiktoken encoding '{encoding}'"
        ) from e


def encode(
    string: str, for_model: str = None, encoding: str | tiktoken.Encoding = None
) -> list[int]:
    """Encodes string to LLM tokens"""
    return _resolve_tiktoken_encoding(for_model, encoding).encode(string)


def num_tokens_from_string(
    string: str, for_model: str = None, encoding: str | tiktoken.Encoding = None
) -> int:
    """Returns the number of tokens in a text string."""
    return len(encode(string, for_model=for_model, encoding=encoding))


def fit_to_token_size(
    docs: list[str],
    max_tokens: int,
    min_documents: int = None,
    for_model: str = None,
    encoding: str | tiktoken.Encoding = None,
) -> tuple[list[str], int]:
    """
    Fit the list of documents to the max_tokens size.
    Returns the new list of documents and qty of removed items
    """
    encoding = _resolve_tiktoken_encoding(for_model, encoding)
    tot_size = 0
    for i, doc in enumerate(docs):
        tot_size += num_tokens_from_string(str(doc), encoding=encoding)
        if min_documents and i < min_documents:
            continue
        if tot_size > max_tokens:
            result = docs[:i]
            return result, len(docs) - len(result)
    return docs, 0



"""Message classes for OpenAI Chat API"""

from enum import Enum
from dataclasses import dataclass, field
from typing import ClassVar


class Role(str, Enum):
    SYSTEM = "system"
    USER = "user"
    ASSISTANT = "assistant"

    def __str__(self):
        return self.value


DEFAULT_MESSAGE_ROLE = Role.USER


@dataclass
class Msg:
    role: str = field(default=DEFAULT_MESSAGE_ROLE)
    content: str = field(default="")

    DICT_FACTORY: ClassVar = dict

    def __str__(self):
        return str(self.content)

    def strip(self):
        self.content = self.content.strip()
        return self


@dataclass
class SysMsg(Msg):
    role: str = field(default=Role.SYSTEM, init=False)


@dataclass
class UserMsg(Msg):
    role: str = field(default=Role.USER, init=False)


@dataclass
class AssistantMsg(Msg):
    role: str = field(default=Role.ASSISTANT, init=False)


class PartialMsg(AssistantMsg):
    """A message that is not fully formed yet."""

    class _PartialMsgDict(dict):
        is_partial = True
        """Custom dictionary class to handle additional properties"""

    placeholder = "<|placeholder|>"
    variants_splitter = "<|or|>"

    DICT_FACTORY: ClassVar = _PartialMsgDict

    @staticmethod
    def split_prefix_and_suffixes(content: str):
        parts = content.split(PartialMsg.placeholder)
        prefix = parts[0]
        suffix = parts[1] if len(parts) > 1 else ""
        suffixes = suffix.split(PartialMsg.variants_splitter) if suffix else []
        return prefix, suffixes

    def prefix_and_suffixes(self):
        return self.split_prefix_and_suffixes(self.content)

    def prefix(self):
        prefix, _ = self.prefix_and_suffixes()
        return prefix

    def suffixes(self):
        _, suffixes = self.prefix_and_suffixes()
        return suffixes



import asyncio
import base64
import logging

import openai

from ..configuration import Config, ApiType
from .._prepare_llm_args import prepare_chat_messages, prepare_prompt
from ..types import LLMAsyncFunctionType, LLMFunctionType, BadAIAnswer
from ..wrappers.llm_response_wrapper import LLMResponse
from ..utils import file_link, is_chat_model, is_image_model
from .shared import make_remove_hidden_output, prepare_callbacks
OPENAI_V1_API = True


def _get_chunk_text(chunk, mode_chat_model: bool):
    # Azure API gives first chunk with empty choices
    if len(chunk.choices) == 0:
        return ""
    choice = chunk.choices[0]
    if mode_chat_model:
        if hasattr(choice, "delta"):
            return getattr(choice.delta, "content", "")
        return ""
    return getattr(choice, "text", "")


async def _a_process_streamed_response(
    response,
    callbacks: list[callable],
    chat_model_used: bool,
    hidden_output_begin: str | None = None,
    hidden_output_end: str | None = None,
):
    response_text: str = ""
    hiding: bool = False
    need_to_hide = hidden_output_begin and hidden_output_end
    async for chunk in response:
        if text_chunk := _get_chunk_text(chunk, chat_model_used):
            if need_to_hide:
                if text_chunk == hidden_output_begin:
                    hiding = True
                    continue
                else:
                    if hiding:
                        if text_chunk == hidden_output_end:
                            hiding = False
                            text_chunk = ""
                        else:
                            continue
            response_text += text_chunk
            for cb in callbacks:
                if asyncio.iscoroutinefunction(cb):
                    await cb(text_chunk)
                else:
                    cb(text_chunk)
    return LLMResponse(response_text, {})


def _process_streamed_response(
    response,
    callbacks: list[callable],
    chat_model_used: bool,
    hidden_output_begin: str | None = None,
    hidden_output_end: str | None = None,
):
    response_text: str = ""
    is_hiding: bool = False
    need_to_hide = hidden_output_begin and hidden_output_end
    for chunk in response:
        if text_chunk := _get_chunk_text(chunk, chat_model_used):
            if need_to_hide:
                if text_chunk == hidden_output_begin:
                    is_hiding = True
                    continue
                else:
                    if is_hiding:
                        if text_chunk == hidden_output_end:
                            is_hiding = False
                            text_chunk = ""
                        else:
                            continue
            response_text += text_chunk
            [cb(text_chunk) for cb in callbacks]
    return LLMResponse(response_text, {})


def _prepare_llm_arguments(config: Config, kwargs: dict):
    args = {**config.LLM_DEFAULT_ARGS, **kwargs}
    args["model"] = args.get(
        "model",
        (
            args.get("deployment_id", config.LLM_DEPLOYMENT_ID or config.MODEL)
            if config.LLM_API_TYPE == ApiType.AZURE
            else config.MODEL
        ),
    )
    callbacks = prepare_callbacks(config, args)
    return args, {"callbacks": callbacks}


def check_for_errors(response):
    if hasattr(response, "object") and response.object == "error":
        raise BadAIAnswer(response.message)
    if hasattr(response, "error") and response.error:
        raise BadAIAnswer(str(response.error))


def make_llm_functions(config: Config) -> tuple[LLMFunctionType, LLMAsyncFunctionType]:
    if config.LLM_API_TYPE == ApiType.AZURE:
        connection_type = openai.AzureOpenAI
        async_connection_type = openai.AsyncAzureOpenAI
        params = {
            "api_key": config.LLM_API_KEY,
            "azure_endpoint": config.LLM_API_BASE,
            "api_version": config.LLM_API_VERSION,
            **config.INIT_PARAMS,
        }
    else:
        connection_type = openai.OpenAI
        async_connection_type = openai.AsyncOpenAI
        params = {
            "api_key": config.LLM_API_KEY,
            "base_url": config.LLM_API_BASE,
            **config.INIT_PARAMS,
        }

    _connection = connection_type(**params)
    _async_connection = async_connection_type(**params)
    remove_hidden_output: callable = make_remove_hidden_output(config)

    async def allm(prompt, **kwargs):
        args, options = _prepare_llm_arguments(config, kwargs)
        if is_chat_model(args["model"], config):
            response = await _async_connection.chat.completions.create(
                messages=prepare_chat_messages(prompt), **args
            )
            check_for_errors(response)
            if args["stream"]:
                return await _a_process_streamed_response(
                    response,
                    options["callbacks"],
                    chat_model_used=True,
                    hidden_output_begin=config.HIDDEN_OUTPUT_BEGIN,
                    hidden_output_end=config.HIDDEN_OUTPUT_END,
                )
            response_text = response.choices[0].message.content
            if config.hiding_output():
                response_text = remove_hidden_output(response_text)
            for cb in options["callbacks"]:
                if asyncio.iscoroutinefunction(cb):
                    await cb(response_text)
                else:
                    cb(response_text)
            return LLMResponse(response_text, response.__dict__)

        response = await _async_connection.completions.create(
            prompt=prepare_prompt(prompt), **args
        )
        check_for_errors(response)
        if args["stream"]:
            return await _a_process_streamed_response(
                response, options["callbacks"], chat_model_used=False
            )
        return LLMResponse(response.choices[0].text, response.__dict__)

    def llm(prompt, **kwargs):
        args, options = _prepare_llm_arguments(config, kwargs)
        if is_image_model(args["model"]):
            save: bool = args.pop("save", True)
            args.pop("stream", None)
            if args["model"] in ["dall-e-2", "dall-e-3"] and "response_format" not in args:
                args["response_format"] = "b64_json"
            if save and args.get("response_format", "b64_json") != "b64_json":
                raise ValueError("Can only save images with response_format='b64_json'")
            response = _connection.images.generate(prompt=prompt, **args)
            check_for_errors(response)
            response_attrs = response.__dict__.copy()
            img_repr = "<image>"
            if save:
                from ..file_storage import storage
                file_name = save if isinstance(save, str) else "generated_images/image.png"
                response_attrs["files"] = []
                for i, img in enumerate(response.data):
                    image_bytes = base64.b64decode(img.b64_json)
                    actual_fn = storage.write(file_name, image_bytes, rewrite_existing=False)
                    actual_fn = storage.abs_path(actual_fn)
                    logging.info(f"Image saved to {file_link(actual_fn)}")
                    response_attrs["files"].append(actual_fn)
                response_attrs["file"] = (
                    response_attrs["files"][0] if response_attrs["files"]
                    else None
                )
                if len(response_attrs["files"]) == 1:
                    img_repr = file_link(response_attrs['file'])
                elif len(response_attrs["files"]) > 1:
                    img_repr = "\n".join(response_attrs["files"])
            for cb in options["callbacks"]:
                cb(img_repr)
            return LLMResponse(img_repr, response_attrs)
        if is_chat_model(args["model"], config):
            response = _connection.chat.completions.create(
                messages=prepare_chat_messages(prompt), **args
            )
            check_for_errors(response)
            if args["stream"]:
                return _process_streamed_response(
                    response,
                    options["callbacks"],
                    chat_model_used=True,
                    hidden_output_begin=config.HIDDEN_OUTPUT_BEGIN,
                    hidden_output_end=config.HIDDEN_OUTPUT_END,
                )
            response_text = response.choices[0].message.content

            if config.hiding_output():
                response_text = remove_hidden_output(response_text)
            for cb in options["callbacks"]:
                cb(response_text)
            return LLMResponse(response_text, response.__dict__)

        # Else (if it is text completion model)
        response = _connection.completions.create(prompt=prepare_prompt(prompt), **args)
        check_for_errors(response)
        if args["stream"]:
            return _process_streamed_response(
                response, options["callbacks"], chat_model_used=False
            )
        return LLMResponse(response.choices[0].text, response.__dict__)

    return llm, allm



import os
from pathlib import Path

import jinja2
from ..types import TplFunctionType


def make_jinja2_env(env) -> jinja2.Environment:
    j2 = jinja2.Environment(
        autoescape=env.config.JINJA2_AUTO_ESCAPE,
        loader=jinja2.ChoiceLoader([
            jinja2.FileSystemLoader(env.config.PROMPT_TEMPLATES_PATH),
            jinja2.FileSystemLoader(Path(__file__).parent.parent / "ai_func"),
        ]),
    )
    j2.globals.update(
        env=env,
        config=env.config,
        **env.config.JINJA2_GLOBALS,
    )
    return j2


def make_tpl_function(env) -> TplFunctionType:
    def tpl(file: os.PathLike[str] | str, **kwargs) -> str:
        return env.jinja_env.get_template(file).render(**kwargs)

    return tpl



import asyncio
import inspect
import threading
from typing import Awaitable, Optional, Any, TypeVar

from ..configuration import Config, LLMConfigError
from .._prepare_llm_args import prepare_chat_messages, prepare_prompt
from ..types import LLMAsyncFunctionType, LLMFunctionType
from ..utils import resolve_callable
from ..wrappers.llm_response_wrapper import LLMResponse
from .shared import prepare_callbacks

T = TypeVar("T")


class _sync_await:
    def __init__(self):
        self._loop = None
        self._looper = None

    def __enter__(self) -> "_sync_await":
        self._loop = asyncio.new_event_loop()
        self._looper = threading.Thread(target=self._loop.run_forever, daemon=True)
        self._looper.start()
        return self

    def __call__(self, coro: Awaitable[T], timeout: Optional[float] = None) -> T:
        return asyncio.run_coroutine_threadsafe(coro, self._loop).result(timeout)

    def __exit__(self, *exc_info: Any) -> None:
        self._loop.call_soon_threadsafe(self._loop.stop)
        self._looper.join()
        self._loop.close()


def _prepare_llm_arguments(config: Config, kwargs: dict):
    args = {**config.LLM_DEFAULT_ARGS, **kwargs}
    args.pop("model", None)
    callbacks = prepare_callbacks(config, args, set_stream=False)
    return args, {"callbacks": callbacks}


def make_llm_functions(
    config: Config, overriden_inference_func: callable = None
) -> tuple[LLMFunctionType, LLMAsyncFunctionType]:
    try:
        inference_fn = resolve_callable(
            overriden_inference_func or config.INFERENCE_FUNC
        )
    except ValueError as e:
        raise LLMConfigError(f"Invalid inference function, {e}") from e

    if inspect.iscoroutinefunction(inference_fn):

        async def allm(prompt, **kwargs):
            args, options = _prepare_llm_arguments(config, kwargs)
            prompt = (
                prepare_chat_messages(prompt)
                if config.CHAT_MODE
                else prepare_prompt(prompt)
            )
            response = await inference_fn(prompt, **args)
            for cb in options["callbacks"]:
                cb(response)
            if not isinstance(response, LLMResponse):
                response = LLMResponse(response)
            return response

        def llm(prompt, **kwargs):
            try:
                loop = asyncio.get_running_loop()
                if loop.is_running():
                    with _sync_await() as sa:
                        return sa(allm(prompt, **kwargs))
                else:
                    # not tested
                    return loop.run_until_complete(allm(prompt, **kwargs))
            except RuntimeError:
                return asyncio.run(allm(prompt, **kwargs))

    else:

        def llm(prompt, **kwargs):
            args, options = _prepare_llm_arguments(config, kwargs)
            prompt = (
                prepare_chat_messages(prompt)
                if config.CHAT_MODE
                else prepare_prompt(prompt)
            )
            response = inference_fn(prompt, **args)
            for cb in options["callbacks"]:
                cb(response)
            if not isinstance(response, LLMResponse):
                response = LLMResponse(response)
            return response

        async def allm(prompt, **kwargs):
            return llm(prompt, **kwargs)

    return llm, allm



import os
from dataclasses import dataclass, asdict
from datetime import datetime
import aiohttp
from .._env import env


@dataclass
class TTSArgs:
    text: str
    out_file: str = None
    voice: str = "D38z5RcWu1voky8WS1ja"
    stability: float = 0.29
    similarity_boost: float = 0.5
    style: float = 0.0
    chunk_size: int = 1024
    speed: float = 1.0
    use_speaker_boost: bool = False
    previous_text: str = None
    next_text: str = None

    def to_dict(self) -> dict:
        return asdict(self)


async def text_to_speech(  # pylint: disable=R0914
    text: str,
    out_file: str = None,
    voice: str = "D38z5RcWu1voky8WS1ja",
    stability=0.29,
    similarity_boost=0.5,
    style=0.0,
    chunk_size=1024,
    speed=1.0,
    use_speaker_boost: bool = False,
    previous_text: str = None,
    next_text: str = None,
) -> str:
    url = f"https://api.elevenlabs.io/v1/text-to-speech/{voice}"
    if not out_file:
        base_name = datetime.now().strftime("%Y-%m-%d_%H_%M_%S__%f")
        os.makedirs(env().config.TEXT_TO_SPEECH_PATH, exist_ok=True)
        out_file = f"{env().config.TEXT_TO_SPEECH_PATH}/{base_name}.mp3"
    data = {
        "text": text,
        "model_id": "eleven_multilingual_v2",
        "voice_settings": {
            "stability": stability,
            "similarity_boost": similarity_boost,
            "style": style,
            "speed": speed,
        },
    }
    if use_speaker_boost:
        data["voice_settings"]["use_speaker_boost"] = use_speaker_boost
    if previous_text:
        data["previous_text"] = previous_text
    if next_text:
        data["next_text"] = next_text
    headers = {
        "Accept": "audio/mpeg",
        "Content-Type": "application/json",
        "xi-api-key": env().config.ELEVENLABS_API_KEY,
    }
    async with aiohttp.request("POST", url, json=data, headers=headers) as response:
        if response.status != 200:
            r = await response.json()
            raise RuntimeError(f"Bad response status: {response.status}: {r}")
        with open(out_file, "wb") as file:
            async for chunk in response.content.iter_chunked(chunk_size):
                if chunk:
                    file.write(chunk)
        return out_file



import time

from .wrappers.llm_response_wrapper import LLMResponse
from ._env import env


class Metrics:
    def __init__(self):
        self._start: float = 0
        self.exec_duration: float = 0
        self.total_gen_duration: float = 0
        self.requests_count: int = 0
        self.succ_requests_count: int = 0
        self.gen_chars_count: int = 0
        self.avg_gen_duration: float = 0
        self.gen_chars_speed: float = 0

    def __enter__(self):
        self._start = time.time()

        env().llm_before_handlers.append(self._before_llm)
        env().llm_after_handlers.append(self._after_llm)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.exec_duration = time.time() - self._start
        env().llm_before_handlers.remove(self._before_llm)
        env().llm_after_handlers.remove(self._after_llm)

    def _before_llm(self, prompt, **kwargs):  # pylint: disable=unused-argument
        self.requests_count += 1

    def _after_llm(self, response: str | LLMResponse):
        self.succ_requests_count += 1
        self.gen_chars_count += len(response) if isinstance(response, str) else 0
        self.total_gen_duration += (
            response.gen_duration if isinstance(response, LLMResponse) else 0
        )
        self.avg_gen_duration = self.total_gen_duration / self.succ_requests_count
        self.gen_chars_speed = (self.gen_chars_count or 1) / (
            self.total_gen_duration or 1
        )



from colorama import Fore, init
from .utils import is_notebook

if not is_notebook():
    init(strip=False)


def info(*args, color=Fore.LIGHTYELLOW_EX, **kwargs):
    print(*[color + str(i) for i in args], **kwargs)


def debug(msg):
    info(msg, color=Fore.BLUE)


def error(*args, **kwargs):
    print(*[Fore.RED + str(i) for i in args], **kwargs)


def warning(*args, **kwargs):
    print(*[Fore.YELLOW + str(i) for i in args], **kwargs)


def ask_yn(msg: str, default: bool | None = None) -> bool:
    """
    Ask user a yes/no question via input() and return their answer as a boolean.
    If default is None, force user to answer y/n with retries.
    If default is True/False, use it on incorrect input or Ctrl-C.
    """
    try:
        input_val = input(msg + " (y/n) ").lower().strip()
        if any(i in input_val for i in ["y", "si", "так", "да", "1", "+"]):
            return True
        if any(i in input_val for i in ["n", "0", "-", "н"]):
            return False
        if default is not None:
            warning("Incorrect input, using default:", "Yes" if default else "No")
            return bool(default)
        warning("Please answer with y/n")
        return ask_yn(msg, default)
    except KeyboardInterrupt:
        warning("Interrupted, using default:", "Yes" if default else "No")
        return bool(default)


def ask_choose(msg: str, variants: list):
    idx = 0
    if isinstance(variants, list):
        for item in variants:
            idx += 1
            print(f"\t{Fore.MAGENTA}{idx}:{Fore.RESET}\t{item}")
    while True:
        i = input(f"{msg} {Fore.MAGENTA}[1-{len(variants)}]{Fore.RESET}: ").strip()
        if not i.isdigit():
            error("Not a number")
            continue
        i = int(i) - 1
        if i >= len(variants) or i < 0:
            error("Incorrect number")
            continue
        break

    item = variants[i]
    return item


def ask_non_empty(msg) -> str:
    while True:
        i = input(msg)
        if i.strip():
            break
        error("Empty input")
    return i


class _ColorFunc(str):
    """
    Cli output coloring function
    """
    def __new__(cls, code):
        obj = str.__new__(cls, code)
        obj.code = code
        return obj

    def __call__(self, *args):
        return f"{self.code}{''.join([str(i) for i in args])}{Fore.RESET}"


# Define colors
red = _ColorFunc(Fore.RED)
green = _ColorFunc(Fore.GREEN)
blue = _ColorFunc(Fore.BLUE)
cyan = _ColorFunc(Fore.CYAN)
yellow = _ColorFunc(Fore.YELLOW)
magenta = _ColorFunc(Fore.MAGENTA)
white = _ColorFunc(Fore.WHITE)
gray = _ColorFunc(Fore.LIGHTBLACK_EX)
reset = _ColorFunc(Fore.RESET)



import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass

import tiktoken

from ..utils import ExtendedString


INT32_MAX = 2**31 - 1  # 2147483647


class SearchResults(list):
    def fit_to_token_size(
        self,
        max_tokens: int,
        min_documents: int = None,
        for_model: str = None,
        encoding: str | tiktoken.Encoding = None,
        verbose=True,
    ):
        from ..tokenizing import fit_to_token_size

        records, removed = fit_to_token_size(
            self,
            max_tokens=max_tokens,
            min_documents=min_documents,
            for_model=for_model,
            encoding=encoding,
        )
        if verbose and removed:
            logging.info(
                "For fitting %d records to %d tokens, %d records was removed",
                len(self),
                max_tokens,
                removed,
            )
        return SearchResults(list(records))


class SearchResult(ExtendedString):
    """
    String containing the search result with additional information in attributes

    Attributes:
        id (str): document (text) identifier in embedding database
        distance (float): The distance between the query and the search result
        metadata (dict): A dictionary containing document metadata
    """

    id: str
    distance: float
    metadata: dict


@dataclass
class AbstractEmbeddingDB(ABC):
    """
    Base class for embedding databases
    """

    @abstractmethod
    def search(
        self,
        collection: str,
        query: str | list,
        n_results: int = 5,
        where: dict = None,
        **kwargs,
    ) -> list[str | SearchResult]:
        """
        Similarity search

        Args:
            collection (str): collection name
            query (str | list): query string or list of query strings
            n_results (int): number of results to return
            where (dict): filter results by metadata
            **kwargs: additional arguments
        """

    def find(self, *args, **kwargs) -> SearchResults | list[str | SearchResult]:
        """
        Alias for `search`
        """
        return self.search(*args, **kwargs)

    @abstractmethod
    def get(
        self,
        collection: str,
        ids: list[str] | str = None,
        limit: int = None,
        offset: int = None,
        where: dict = None,
        **kwargs,
    ) -> list[str | SearchResult] | str | SearchResult | None:
        """
        Get documents

        Args:
            collection (str): collection name
            ids (list[str] | str): document id or list of document ids
            limit (int): maximum number of documents to return
            offset (int): number of documents to skip
            where (dict): filter results by metadata

        Returns:
            List of documents or single document
        """

    def find_all(
        self,
        collection: str,
        query: str | list,
        where: dict = None,
        **kwargs,
    ) -> SearchResults | list[str | SearchResult]:
        return self.search(
            collection, query, n_results=INT32_MAX, where=where, **kwargs
        )

    @abstractmethod
    def get_all(self, collection: str) -> SearchResults | list[str | SearchResult]:
        """Return all documents in the collection"""

    def save(self, collection: str, text: str, metadata: dict = None):
        """Save a single document in the collection"""
        self.save_many(collection, [(text, metadata)])

    @abstractmethod
    def save_many(self, collection: str, items: list[tuple[str, dict] | str]):
        """Save multiple documents in the collection"""

    @abstractmethod
    def clear(self, collection: str):
        """Clear the collection"""

    def find_one(self, collection: str, query: str | list) -> str | SearchResult | None:
        """
        Find most similar document in the collection

        Returns:
            Most similar document or None if collection is empty
        """
        return next(iter(self.search(collection, query, 1)), None)

    @abstractmethod
    def count(self, collection: str) -> int:
        """
        Count the number of documents in the collection

        Returns:
            Number of documents in the collection, 0 if collection does not exist
        """

    @abstractmethod
    def delete(self, collection: str, what: str | list[str] | dict):
        """
        Delete documents from the collection

        Args:
            collection (str): collection name
            what (str | list[str] | dict): id, list ids or metadata query
        """

    @abstractmethod
    def collection_exists(self, collection: str) -> bool:
        """
        Check if the collection exists

        Returns:
            True if collection exists, False otherwise
        """

    def has_content(self, collection: str) -> bool:
        """
        Check if the collection exists anf contains any documents

        Returns:
            True if collection exists and contains documents, False otherwise
        """
        return self.count(collection) > 0



from typing import TYPE_CHECKING, Union

from .._llm_functions import allm, llm
from ..utils import ExtendedString, ConvertableToMessage

if TYPE_CHECKING:
    from .llm_response_wrapper import LLMResponse  # noqa: F401


class PromptWrapper(ExtendedString, ConvertableToMessage):
    """
    A utility class that wraps a prompt string, extending it with convenient methods
    for enhanced functionality.
    """

    def to_llm(self, **kwargs) -> Union[str, "LLMResponse"]:
        """
        Send prompt to Large Language Model, see `llm`
        """
        return llm(self, **kwargs)

    async def to_allm(self, **kwargs) -> Union[str, "LLMResponse"]:
        """
        Send prompt to Large Language Model asynchronously, see `allm`
        """
        return await allm(self, **kwargs)



import logging
import gc

import transformers
import torch

from .local_llm import make_llm_functions as make_local_llm_functions
from ..message_types import PartialMsg
from ..wrappers.llm_response_wrapper import LLMResponse
from ..configuration import Config, LLMConfigError
from ..types import LLMFunctionType, LLMAsyncFunctionType


def inference(prompt: str, model, tokenizer, **kwargs):
    skip_special_tokens = kwargs.pop("skip_special_tokens", True)
    inputs = tokenizer(prompt, return_tensors="pt").to(model.device)
    outputs = model.generate(**inputs, **kwargs)
    outputs = [
        tokenizer.decode(i[len(inputs[0]):], skip_special_tokens=skip_special_tokens)
        for i in outputs
    ]
    return LLMResponse(outputs[0], dict(all=outputs))


def pipeline_inference(prompt: str, pipeline, **kwargs):
    raw_output = pipeline(prompt, **{"return_full_text": False, **kwargs})
    return LLMResponse(
        raw_output[0]["generated_text"],
        dict(all=list(map(lambda x: x["generated_text"], raw_output))),
    )


def clear_mem():
    gc.collect()
    torch.cuda.empty_cache()


def try_make_model_config(config: Config):
    try:
        model_config = transformers.AutoConfig.from_pretrained(
            config.MODEL,
            trust_remote_code=True,
        )
    except OSError:
        model_config = None
    if config.INIT_PARAMS.get("gradient_checkpointing"):
        if model_config:
            model_config.gradient_checkpointing = True
        else:
            raise LLMConfigError(
                "Can't apply gradient checkpointing without model config."
            )
    return model_config


def get_last_message_ending(tokenizer) -> str:
    try:
        sample_content = "sample"
        sample_prompt = tokenizer.apply_chat_template(
            [
                dict(role="user", content="hi"),
                dict(role="assistant", content=sample_content),
            ],
            add_generation_prompt=True,
            tokenize=False,
        )
        pos = sample_prompt.rfind(sample_content) + len(sample_content)
        return sample_prompt[pos:]
    except:  # pylint: disable=bare-except # noqa
        logging.warning(
            "Can't determine the ending of the last message in the tokenizer's chat template."
        )
        return ""


def resolve_tokenizer(config: Config):
    return config.INIT_PARAMS.get(
        "tokenizer"
    ) or transformers.AutoTokenizer.from_pretrained(
        config.MODEL, trust_remote_code=True
    )


def resolve_model(config: Config):
    params = config.INIT_PARAMS
    if not (model := params.get("model")):
        clear_mem()
        mc_param_names = [
            "model",  # custom transformers model instance
            "tokenizer",  # custom tokenizer instance
            "device",  # device if model.to(device) is needed
            "quantize_4bit",  # bool, default quantization config will be used
            "inference",  # callable, custom inference function.
            # this is different from Config.INFERENCE_FUNC,
            # because accepts model and tokenizer as arguments
            "always_clear_mem",  # gc.collect() & torch.cuda.empty_cache() before inference
            "gradient_checkpointing",  # bool, enable gradient checkpointing
            "use_pipeline",  # bool
            "pipeline_task",  # str, pipeline task name
            "stops_from_chat_template",  # bool, use chat template to determine stopping criteria
        ]

        model_init_params = {
            **dict(
                trust_remote_code=True,
                dtype="auto",
                device_map="auto",
                offload_folder=config.STORAGE_PATH,
            ),
            **{k: v for k, v in params.items() if k not in mc_param_names},
        }
        if "config" not in model_init_params:
            model_init_params["config"] = try_make_model_config(config)

        if (
            params.get("quantize_4bit")
            and "quantization_config" not in model_init_params
        ):
            model_init_params["quantization_config"] = transformers.BitsAndBytesConfig(
                load_in_4bit=True,
                bnb_4bit_quant_type="nf4",
                bnb_4bit_compute_dtype=torch.bfloat16,
                bnb_4bit_use_double_quant=True,
            )

        model = transformers.AutoModelForCausalLM.from_pretrained(
            config.MODEL, **model_init_params
        )
        try:
            model.generation_config = transformers.GenerationConfig.from_pretrained(
                config.MODEL
            )
        except:  # pylint: disable=bare-except  # noqa
            logging.warning("Can't create generation config")

        if "device" in params:
            model.to(params["device"])

    return model


def make_llm_functions(
    config: Config, env
) -> tuple[LLMFunctionType, LLMAsyncFunctionType]:
    logging.info(f"Loading local Transformers model {config.MODEL}...")
    params = config.INIT_PARAMS

    tokenizer = resolve_tokenizer(config)
    model = resolve_model(config)

    transformers_model_args = {
        "max_new_tokens": 2048,
    }
    if hasattr(tokenizer, "eos_token_id") and tokenizer.eos_token_id:
        transformers_model_args["eos_token_id"] = tokenizer.eos_token_id

    inference_fn = params.get("inference") or inference
    if isinstance(inference_fn, str):
        inference_fn = globals()[inference_fn]
    use_pipeline = params.get("use_pipeline", False)
    if use_pipeline:
        pipeline = transformers.pipeline(
            params.get("pipeline_task", "text-generation"),
            model=model,
            tokenizer=tokenizer,
            dtype=params.get("torch_dtype") or params.get("dtype") or "auto",
            device_map=params.get("device_map", "auto"),
        )
        setattr(env, "pipeline", pipeline)

    def make_stopping_criteria(seq: str | list[str]) -> list[callable]:
        if isinstance(seq, list):
            funcs = []
            for s in seq:
                funcs += make_stopping_criteria(s)
            return funcs

        stop_seq = tokenizer.encode(
            seq, add_special_tokens=False, return_tensors="pt"
        ).to(model.device)[0]
        # Some tokenizers like Phi-3 (SentencePiece tokenizer) add an extra token at the beginning
        # independently of add_special_tokens=False
        if stop_seq[0] == 29871:  # SENTENCEPIECE_UNDERLINE
            stop_seq = stop_seq[1:]
        offset = -len(stop_seq)

        def stopping_criteria(ids, *args):  # pylint: disable=unused-argument
            return torch.all(torch.eq(ids[0][offset:], stop_seq))

        return [stopping_criteria]

    setattr(env, "tokenizer", tokenizer)
    setattr(env, "model", model)
    setattr(env, "inference", inference_fn)
    setattr(env, "make_stopping_criteria", make_stopping_criteria)
    always_clear_mem = params.get("always_clear_mem", False)
    ending = get_last_message_ending(tokenizer) if config.CHAT_MODE else ""

    stops_from_chat_template: list[callable] = []
    if params.get("stops_from_chat_template", True) and ending:
        stops_from_chat_template: list[callable] = make_stopping_criteria(
            ending.strip()
        )

    def wrapped_inference(prompt: list[dict] | str, **kwargs):
        if (n := kwargs.pop("n", None)) is not None:  # open_ai style
            kwargs["num_return_sequences"] = n

        if (seed := kwargs.pop("seed", None)) is not None:  # open_ai style
            transformers.set_seed(seed)

        if (stop := kwargs.pop("stop", None)) is not None:  # open_ai style
            kwargs["stopping_criteria"] = make_stopping_criteria(stop)

        partial_msg_used = False
        if config.CHAT_MODE:
            sc = kwargs.get("stopping_criteria", [])
            if stops_from_chat_template:
                sc += stops_from_chat_template
            if isinstance(prompt, list) and len(prompt) > 0:
                last = prompt[-1]
                if isinstance(last, dict) and hasattr(last, "is_partial"):
                    kwargs["continue_last_message"] = True
                    prefix, suffixes = PartialMsg.split_prefix_and_suffixes(
                        last["content"]
                    )
                    last["content"] = prefix
                    if suffixes:
                        sc += make_stopping_criteria(suffixes)
                    partial_msg_used = True
            if sc:
                kwargs["stopping_criteria"] = sc
            prompt = tokenizer.apply_chat_template(
                prompt, add_generation_prompt=True, tokenize=False
            )
            if kwargs.pop("continue_last_message", False) and ending:
                prompt = prompt[: -len(ending)]
        if seed := kwargs.pop("seed", None):
            torch.manual_seed(seed)
        args = {**transformers_model_args, **kwargs}
        always_clear_mem and clear_mem()
        if use_pipeline:
            out = pipeline_inference(prompt, pipeline, **args)
        else:
            out = env.inference(
                prompt, model=env.model, tokenizer=env.tokenizer, **args
            )
        if not isinstance(out, LLMResponse):
            out = LLMResponse(out)
        if partial_msg_used:
            # original generation doesn't include the prefix, but includes the suffix
            actual_suffix = ""
            for s in suffixes:
                if out.endswith(s):
                    actual_suffix = s
                    break
            setattr(
                out,
                "inner",
                LLMResponse(out[: -len(actual_suffix)]) if actual_suffix else out,
            )
            setattr(out, "outer", LLMResponse(prefix + out))
            setattr(out, "prefixed", LLMResponse(prefix + out.inner))
        return out

    logging.info(f"Local Transformers model loaded: {config.MODEL}")
    return make_local_llm_functions(config, wrapped_inference)



import logging
import asyncio
import anthropic
from anthropic.types import ContentBlockDeltaEvent

from ..configuration import Config
from .._prepare_llm_args import prepare_chat_messages
from ..message_types import Role
from ..types import LLMAsyncFunctionType, LLMFunctionType
from ..wrappers.llm_response_wrapper import LLMResponse
from .shared import prepare_callbacks


def _get_chunk_text(chunk):
    return isinstance(chunk, ContentBlockDeltaEvent) and chunk.delta.text or ""


async def _a_process_streamed_response(response, callbacks: list[callable]):
    response_text: str = ""
    async for chunk in response:
        if text_chunk := _get_chunk_text(chunk):
            response_text += text_chunk
            for cb in callbacks:
                if asyncio.iscoroutinefunction(cb):
                    await cb(text_chunk)
                else:
                    cb(text_chunk)
    return LLMResponse(response_text, {})


def _process_streamed_response(response, callbacks: list[callable]):
    response_text: str = ""
    for chunk in response:
        if text_chunk := _get_chunk_text(chunk):
            response_text += text_chunk
            [cb(text_chunk) for cb in callbacks]
    return LLMResponse(response_text, {})


def _prepare_llm_arguments(config: Config, kwargs: dict):
    args = {**config.LLM_DEFAULT_ARGS, **kwargs}
    args["model"] = args.get("model", config.MODEL)
    if "max_tokens" not in args:
        if "claude-3-5-sonnet" in args["model"]:
            args["max_tokens"] = 8192
        elif "claude-3-7-sonnet" in args["model"]:
            args["max_tokens"] = 16384
        else:
            args["max_tokens"] = 4096
    # Remove arguments not supported by Anthropic
    args.pop("seed", None)
    args.pop("n", None)
    if "temperature" in args and "top_p" in args:
        del args["top_p"]
        logging.warning(
            "`temperature` and `top_p` cannot both be specified for this model. "
            "`top_p` parameter will be ignored. "
        )
    callbacks = prepare_callbacks(config, args)
    return args, {"callbacks": callbacks}


def _extract_sys_msg(prepared_messages: list[dict]) -> tuple[str, list[dict]]:
    """
    Anthropic does not support system messages,
    so we need to extract them to pass as a separate argument.
    Also ensures the first and last messages are from user,
    and there is assistants message between user messages.
    """
    system = "\n".join(
        i["content"] for i in prepared_messages if i.get("role") == Role.SYSTEM
    )
    messages = [i for i in prepared_messages if i.get("role") != Role.SYSTEM]

    empty_user_msg = {"role": Role.USER, "content": "--//--"}
    if not messages or messages[0]["role"] != Role.USER:
        messages.insert(0, empty_user_msg)

    # Ensure proper alternation and last message is from User
    normalized_messages = []
    expected_role = Role.USER
    for msg in messages:
        if msg["role"] == expected_role:
            normalized_messages.append(msg)
            expected_role = Role.ASSISTANT if expected_role == Role.USER else Role.USER
        elif msg["role"] == Role.USER and expected_role == Role.ASSISTANT:
            normalized_messages.append({"role": Role.ASSISTANT, "content": "--//--"})
            normalized_messages.append(msg)
            expected_role = Role.ASSISTANT

    # Ensure the last message is from User
    if normalized_messages[-1]["role"] != Role.USER:
        normalized_messages.append(empty_user_msg)

    return system, normalized_messages


def make_llm_functions(config: Config) -> tuple[LLMFunctionType, LLMAsyncFunctionType]:
    sync_client = anthropic.Anthropic(
        api_key=config.LLM_API_KEY,
        base_url=config.LLM_API_BASE,
        **config.INIT_PARAMS,
    )
    async_client = anthropic.AsyncAnthropic(
        api_key=config.LLM_API_KEY,
        base_url=config.LLM_API_BASE,
        **config.INIT_PARAMS,
    )

    async def allm(prompt, **kwargs):
        args, options = _prepare_llm_arguments(config, kwargs)
        args["system"], args["messages"] = _extract_sys_msg(
            prepare_chat_messages(prompt)
        )
        response = await async_client.messages.create(**args)
        if args.get("stream"):
            return await _a_process_streamed_response(response, options["callbacks"])

        for cb in options["callbacks"]:
            if asyncio.iscoroutinefunction(cb):
                await cb(response.content[0].text)
            else:
                cb(response.content[0].text)
        return LLMResponse(response.content[0].text, response.__dict__)

    def llm(prompt, **kwargs):
        args, options = _prepare_llm_arguments(config, kwargs)
        args["system"], args["messages"] = _extract_sys_msg(
            prepare_chat_messages(prompt)
        )
        response = sync_client.messages.create(**args)
        if args.get("stream"):
            return _process_streamed_response(response, options["callbacks"])

        for cb in options["callbacks"]:
            cb(response.content[0].text)
        return LLMResponse(response.content[0].text, response.__dict__)

    return llm, allm



"""
# Minimalistic Foundation for AI Applications

**microcore** is a collection of python adapters for Large Language Models
and Semantic Search APIs allowing to
communicate with these services convenient way, make it easily switchable
and separate business logic from implementation details.
"""

import os
from . import mcp
from . import ui
from . import tokenizing
from . import presets
from .embedding_db import SearchResult, AbstractEmbeddingDB, SearchResults
from .file_storage import storage
from ._env import configure, env, config, min_setup
from .logging import use_logging
from .message_types import UserMsg, AssistantMsg, SysMsg, Msg, PartialMsg
from .configuration import (
    ApiType,
    LLMApiBaseError,
    LLMApiDeploymentIdError,
    LLMApiKeyError,
    LLMApiVersionError,
    LLMConfigError,
    Config,
    EmbeddingDbType,
    PRINT_STREAM,
)
from .types import BadAIJsonAnswer, BadAIAnswer, LLMContextLengthExceededError
from .wrappers.prompt_wrapper import PromptWrapper
from .wrappers.llm_response_wrapper import LLMResponse
from ._llm_functions import llm, allm, llm_parallel
from .utils import parse, dedent
from .metrics import Metrics
from .interactive_setup import interactive_setup


def tpl(
    file: os.PathLike[str] | str,
    sanitize_utf8=True,
    **kwargs
) -> str | PromptWrapper:
    """Renders a prompt template using the provided parameters."""
    rendered = env().tpl_function(file, **kwargs)
    if sanitize_utf8:
        rendered = rendered.encode('utf-8', errors='replace').decode('utf-8')
    return PromptWrapper(rendered, kwargs)


def prompt(
    template_str: str,
    remove_indent=True,
    sanitize_utf8=True,
    **kwargs
) -> str | PromptWrapper:
    """Renders a prompt template from string using the provided parameters."""
    if remove_indent:
        template_str = dedent(template_str)
    rendered = env().jinja_env.from_string(template_str).render(**kwargs)
    if sanitize_utf8:
        rendered = rendered.encode('utf-8', errors='replace').decode('utf-8')
    return PromptWrapper(rendered, kwargs)


fmt = prompt


def use_model(name: str):
    """Switches language model"""
    config().MODEL = name
    config().LLM_DEFAULT_ARGS["model"] = name


def validate_config():
    """
    Validates current MicroCore configuration

    Raises:
        `LLMConfigError` if configuration is invalid
    """
    config().validate()


class _EmbeddingDBProxy(AbstractEmbeddingDB):
    def get_all(self, collection: str) -> list[str | SearchResult]:
        return env().texts.get_all(collection)

    def search(
        self,
        collection: str,
        query: str | list,
        n_results: int = 5,
        where: dict = None,
        **kwargs,
    ) -> SearchResults | list[str | SearchResult]:
        return env().texts.search(collection, query, n_results, where, **kwargs)

    def find(self, *args, **kwargs) -> SearchResults | list[str | SearchResult]:
        return self.search(*args, **kwargs)

    def find_all(
        self,
        collection: str,
        query: str | list,
        where: dict = None,
        **kwargs,
    ) -> SearchResults | list[str | SearchResult]:
        return env().texts.find_all(collection, query, where, **kwargs)

    def get(
        self,
        collection: str,
        ids: list[str] | str = None,
        limit: int = None,
        offset: int = None,
        where: dict = None,
        **kwargs,
    ) -> list[str | SearchResult] | str | SearchResult | None:
        return env().texts.get(collection, ids, limit, offset, where, **kwargs)

    def save_many(self, collection: str, items: list[tuple[str, dict] | str]):
        return env().texts.save_many(collection, items)

    def save(self, collection: str, text: str, metadata: dict = None):
        return env().texts.save(collection, text, metadata)

    def clear(self, collection: str):
        return env().texts.clear(collection)

    def count(self, collection: str) -> int:
        return env().texts.count(collection)

    def delete(self, collection: str, what: str | list[str] | dict):
        return env().texts.delete(collection, what)

    def collection_exists(self, collection: str) -> bool:
        return env().texts.collection_exists(collection)

    def has_content(self, collection: str) -> bool:
        return env().texts.has_content(collection)


texts = _EmbeddingDBProxy()
"""Embedding database, see `microcore.embedding_db.AbstractEmbeddingDB`"""


def mcp_server(name: str) -> mcp.MCPServer:  # noqa, pylint-disable=E0602
    """
    Returns MCP server by name from the registry.

    Args:
        name (str): The name of the MCP server.

    Returns:
        MCPServer: The MCP server instance.

    Raises:
        ValueError: If the server with the given name is not found in the registry.
    """
    return mcp.server(name)  # noqa, pylint-disable=E0602


__all__ = [
    "llm",
    "allm",
    "llm_parallel",
    "tpl",
    "prompt",
    "fmt",
    "texts",
    "configure",
    "min_setup",
    "validate_config",
    "storage",
    "use_model",
    "use_logging",
    "env",
    "config",
    "Msg",
    "UserMsg",
    "SysMsg",
    "AssistantMsg",
    "PartialMsg",
    "ApiType",
    "EmbeddingDbType",
    "BadAIJsonAnswer",
    "PRINT_STREAM",
    "presets",
    "BadAIAnswer",
    "LLMApiBaseError",
    "LLMApiDeploymentIdError",
    "LLMApiKeyError",
    "LLMApiVersionError",
    "LLMConfigError",
    "LLMContextLengthExceededError",
    "LLMResponse",
    "PromptWrapper",
    "parse",
    "SearchResult",
    "SearchResults",
    "dedent",
    # submodules
    "embedding_db",
    "file_storage",
    "message_types",
    "utils",
    "configuration",
    "Config",
    "types",
    "ui",
    "mcp",
    "mcp_server",
    "tokenizing",
    "Metrics",
    "interactive_setup",
    # "wrappers",
]

__version__ = "4.5.0"



import os

from .configuration import EmbeddingDbType, ApiType, Config
from .ui import ask_choose, ask_non_empty, ask_yn, error, yellow
from ._env import configure
from ._llm_functions import llm
from .utils import file_link


def interactive_setup(
    file_path: str,
    defaults: dict = None,
    extras: dict | list = None,
) -> Config | None:
    """
    Start interactive CLI setup for LLM API configuration.
    Prompts user for configuration details such as API type, key, model name,
    and base URL. Tests the LLM API with a sample query and saves the configuration
    to a specified file if the user chooses to do so.
    Args:
        file_path (str): Path to the configuration file.
        defaults (dict, optional): Default configuration values.
            If provided, user will not be prompted for those values.
            Additional values for storing in the file can be added to defaults.
        extras (dict | list, optional): Additional configuration fields to prompt for.
    """
    raw_config = dict(defaults) if defaults else dict()
    if "LLM_API_TYPE" not in raw_config:
        raw_config["LLM_API_TYPE"] = ask_choose(
            "Choose LLM API Type:",
            list(i.value for i in ApiType if not ApiType.is_local(i)),
        )
    if "LLM_API_KEY" not in raw_config:
        raw_config["LLM_API_KEY"] = ask_non_empty("API Key: ").strip()
    if "MODEL" not in raw_config:
        raw_config["MODEL"] = ask_non_empty("Model Name: ").strip()
    if "LLM_API_BASE" not in raw_config:
        raw_config["LLM_API_BASE"] = input(
            "API Base URL (may be empty for some API types): "
        ).strip()
    if extras:
        if isinstance(extras, list):
            extras = {
                i: (
                    str(i)
                    .replace('_', ' ')
                    .capitalize()
                    .replace('Llm', 'LLM')
                    .replace('Api', 'API')
                )
                for i in extras
            }
        for field, title in extras.items():
            if field not in raw_config:
                raw_config[field] = ask_non_empty(f"{title}: ").strip()
    try:
        config_dict = dict(
            USE_DOT_ENV=False,
            EMBEDDING_DB_TYPE=EmbeddingDbType.NONE,
            USE_LOGGING=True,
        )
        config_dict.update(raw_config)
        config = configure(config_dict)
        print("Testing LLM...")
        q = "What is capital of France?\n(!) IMPORTANT: Answer only with one word"
        assert "pari" in llm(q).lower()
    except Exception as e:  # pylint: disable=W0718
        error(f"Error testing LLM API: {e}")
        if ask_yn("Restart configuring?"):
            return interactive_setup(file_path, defaults, extras)
        return None

    config_body = ''.join(f"{k}={v}\n" for k, v in raw_config.items())
    print(f"Configuration:\n{yellow(config_body)}")
    if ask_yn(f"Save configuration to file {file_link(file_path)}?"):
        dir_path = os.path.dirname(file_path)
        if dir_path:
            os.makedirs(dir_path, exist_ok=True)
        with open(file_path, "w", encoding="utf-8") as f:
            f.write(config_body)
        print(f"Saved to {file_link(file_path)}")
    return config



import logging
import random
import subprocess
import sys
from datetime import datetime

from .file_storage import storage
from .utils import file_link, dedent


def make_silent(code: str) -> str:
    definition = "def dummy_print(*args, **kwargs): ...\n"
    code = code.replace("print(", "dummy_print(")
    if definition not in code:
        code = definition + code
    return code


def execute(
    program: str,
    name=None,
    timeout=15,
    cleanup: bool = True,
    traceback: bool = True,
    log_errors: bool = True,
) -> tuple[str, str]:
    """Executes the provided Python program and returns the output and error messages."""
    program = dedent(program)
    if name is None:
        date = datetime.now().strftime("%Y_%m_%d")
        rnd_int = random.randint(0, sys.maxsize - 1)
        name = f"code/program_{date}__{rnd_int}"
    fn = storage.abs_path(storage.write(f"{name}.py", program, rewrite_existing=False))
    logging.info(f"Executing the program: {file_link(fn)}")
    cmd = [sys.executable, fn]
    stdout, stderr = "", ""
    try:
        with subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
        ) as proc:
            try:
                stdout, stderr = proc.communicate(timeout=timeout)
                if proc.returncode != 0:
                    log_errors and logging.error(
                        f"Error executing the program, STDERR:\n<\n{stderr}\n>\n",
                    )
            except subprocess.TimeoutExpired:
                proc.kill()
                error = (
                    f"The command timed out (max execution time is {timeout} seconds)"
                )
                stderr = f"{stderr}\n{error}"
                log_errors and logging.error(stderr)
    finally:
        if cleanup:
            try:
                storage.delete(fn)
            except Exception as e:  # pylint: disable=broad-except
                logging.error(f"Error deleting the file: {fn}", e)
        else:
            logging.info(f"Executed code: {file_link(fn)}")

    stdout, stderr = str(stdout).strip(), str(stderr).strip()

    if stderr and not traceback:
        stderr = stderr.splitlines()[-1].lstrip()

    return stdout, stderr



from typing import Any
from typing import TYPE_CHECKING

from ..types import BadAIAnswer, TPrompt
from ..utils import ExtendedString, ConvertableToMessage, extract_number
from ..message_types import Role, AssistantMsg

if TYPE_CHECKING:
    from ..mcp import MCPConnection, MCPAnswer


class DictFromLLMResponse(dict):
    llm_response: "LLMResponse"

    def from_llm_response(self, llm_response: "LLMResponse"):
        self.llm_response = llm_response
        return self

    async def to_mcp(self, mcp: "MCPConnection"):
        return await mcp.exec(self)


class LLMResponse(ExtendedString, ConvertableToMessage):
    """
    Response from the Large Language Model.

    If treated as a string, it returns the text generated by the LLM.

    Also, it contains all fields returned by the API accessible as an attributes.

    See fields returned by the OpenAI:

    - https://platform.openai.com/docs/api-reference/completions/object
    - https://platform.openai.com/docs/api-reference/chat/object
    """

    role: Role
    content: str
    prompt: TPrompt
    gen_duration: float

    def __new__(cls, string: str, attrs: dict = None):
        attrs = {
            **(attrs or {}),
            "role": Role.ASSISTANT,
            "content": str(string),
            "prompt": None,
            # generation duration in seconds (float), used in metrics
            "gen_duration": None,
        }
        obj = ExtendedString.__new__(cls, string, attrs)
        return obj

    def parse_json(
        self,
        raise_errors: bool = True,
        required_fields: list[str] = None,
        validator: callable = None,
    ) -> list | dict | float | int | str | DictFromLLMResponse:
        try:
            res = super().parse_json(raise_errors=True, required_fields=required_fields)
            if validator:
                try:
                    validator(res)
                except Exception as e:
                    raise BadAIAnswer(f"Language model response validation failed: {e}") from None
        except Exception as e:  # pylint: disable=W0718
            if hasattr(self, "_retry_callback"):
                res = self._retry_callback()
                if isinstance(res, DictFromLLMResponse):
                    return res
                return res.parse_json(raise_errors, required_fields, validator)
            if raise_errors:
                raise e
            res = False
        if isinstance(res, dict):
            res = DictFromLLMResponse(res)
            res.llm_response = self
        return res

    def parse_number(
        self,
        default=BadAIAnswer,
        position="last",
        dtype: type | str = float,
        rounding: bool = False,
    ) -> int | float | Any:
        return extract_number(self.content, default, position, dtype, rounding)

    def as_message(self) -> AssistantMsg:
        return self.as_assistant

    async def to_mcp(self, mcp: "MCPConnection") -> "MCPAnswer":
        return await mcp.exec(self)

    def is_tool_call(self):
        from .._env import env
        return self.parse_json(
            raise_errors=False,
            required_fields=[env().config.AI_SYNTAX_FUNCTION_NAME_FIELD],
        ) is not False



from typing import TYPE_CHECKING, Callable, Any, Awaitable, Union, List
from os import PathLike

from .message_types import Msg

if TYPE_CHECKING:
    from .wrappers.prompt_wrapper import PromptWrapper  # noqa: F401
    from .wrappers.llm_response_wrapper import LLMResponse  # noqa: F401

TPrompt = Union[
    dict, Msg, str, "PromptWrapper", List[Union[dict, Msg, str, "PromptWrapper"]]
]
"""Type for prompt argument in LLM requests"""
TplFunctionType = Callable[[Union[PathLike[str], str], Any], str]
"""Function type for rendering prompt templates"""
LLMFunctionType = Callable[[TPrompt, Any], "LLMResponse"]
"""Function type for requesting LLM synchronously"""
LLMAsyncFunctionType = Callable[[TPrompt, Any], Awaitable["LLMResponse"]]
"""Function type for requesting LLM asynchronously"""


class BadAIAnswer(ValueError):
    """Unprocessable response generated by the LLM"""

    def __init__(self, message: str = None, details: str = None):
        self.message = str(message or "Unprocessable response generated by the LLM")
        self.details = details
        super().__init__(self.message + (f": {self.details}" if self.details else ""))

    def __str__(self):
        return self.message + (f": {self.details}" if self.details else "")


class BadAIJsonAnswer(BadAIAnswer):
    def __init__(
        self, message: str = "Invalid JSON generated by the LLM", details=None
    ):
        super().__init__(message, details)


class LLMContextLengthExceededError(BadAIAnswer):
    def __init__(
        self,
        message: str = None,
        actual_tokens: int | None = None,
        max_tokens: int | None = None,
        model: str | None = None,
    ):
        if message is None:
            message = "LLM context length exceeded" + (f" for model '{model}'" if model else "")
            if actual_tokens:
                message += f", actual tokens: {actual_tokens}"
            if max_tokens:
                message += f", max. tokens: {max_tokens}"
        super().__init__(message)
        self.actual_tokens = actual_tokens
        self.max_tokens = max_tokens
        self.model = model



import json
import re
from enum import Enum

from .types import BadAIJsonAnswer


def safe_remove_outer_json_wrapper(input_string: str) -> str:
    if input_string.endswith("```"):
        if input_string.startswith("```json"):
            return input_string[7:-3].strip()
        if input_string.startswith("```"):
            return input_string[3:-3].strip()
    return input_string


def simple_json_format_check(input_string: str):
    for opener, close in [("{", "}"), ("[", "]"), ('"', '"')]:
        if input_string.startswith(opener) and input_string.endswith(close):
            return True
    if input_string.isdigit() or input_string in ["true", "false", "null"]:
        return True
    try:
        float(input_string)
        return True
    except ValueError:
        ...
    return False


class ExtractStrategy(str, Enum):
    FIRST = "first"
    LAST = "last"
    OUTER = "outer"


def extract_block(
    input_string: str,
    block_begin: str,
    block_end: str,
    include_wrapper: bool = False,
    strategy: ExtractStrategy = ExtractStrategy.OUTER,
) -> str | None:
    """
    Extracts a block of text between two markers from a string.
    """
    assert block_begin and block_end, "Block markers cannot be empty."
    start = -1
    begin_len = 1
    if strategy in (ExtractStrategy.OUTER, ExtractStrategy.FIRST):
        start = input_string.find(block_begin)
        if start == -1:
            return None
        begin_len = len(block_begin)
    if strategy in (ExtractStrategy.OUTER, ExtractStrategy.LAST):
        end = input_string.rfind(block_end, start + begin_len)
    else:  # first
        end = input_string.find(block_end, start + begin_len)
    if end == -1:
        return None
    if strategy == ExtractStrategy.LAST:
        start = input_string.rfind(block_begin, None, end)
        if start == -1:
            return None

    start_capture = start if include_wrapper else start + len(block_begin)
    end_capture = end + len(block_end) if include_wrapper else end

    return input_string[start_capture:end_capture]


def unwrap_json_substring(
    input_string: str, allow_in_text: bool = True, return_original_on_fail: bool = True
) -> str:
    input_string = safe_remove_outer_json_wrapper(str(input_string).strip())
    if simple_json_format_check(input_string):
        return input_string

    if not allow_in_text:
        return input_string if return_original_on_fail else ""

    if (val := extract_block(input_string, "```json", "```", False)) is not None:
        return val.strip()

    # find outermost {} or []
    brace = None
    start = 0
    end = 0
    try:
        start, end = input_string.index("{"), input_string.rindex("}")
        if end > start:
            brace = "{"
    except ValueError:
        ...
    try:
        s_start, s_end = input_string.index("["), input_string.rindex("]")
        if not brace or (s_start < start and s_end > end):
            start = s_start
            end = s_end
            brace = "["
    except ValueError:
        ...

    return (
        input_string[start: end + 1]
        if brace
        else input_string if return_original_on_fail else ""
    )


# pylint: disable=too-many-return-statements
def fix_json(s: str) -> str:
    """
    Fix internal JSON content.
    Note: this function should not be used for valid JSON strings.
    Args:
        s (str): AI-generated JSON string containing errors.
    """

    def between_lines(pattern):
        return (
            rf"({json_obj_before}{pattern}{json_obj_after})"
            rf"|({json_list_before}{pattern}{json_list_after})"
        )

    json_obj_before = (
        r"((?<=\{)|(?<=[\"\d]\,)|"
        r"(?<=null\,)|(?<=true\,)|(?<=false\,)|(?<=[\"\d])|"
        r"(?<=null)|(?<=true)|(?<=false))\s*"
    )
    json_obj_after = r'\s*((?=\})|(?=\"[^"\n]+\"\s*\:\s*))'
    json_list_before = (
        r"((?<=\[)|(?<=[\"\d]\,)|(?<=null\,)|(?<=true\,)|(?<=false\,)|"
        r"(?<=[\"\d])|(?<=null)|(?<=true)|(?<=false))\s*"
    )
    json_list_after = r"\s*((?=[\]\"\d])|(?=true)|(?=false)|(?=null))"
    json_before = rf"({json_obj_before}|{json_list_before})"
    json_after = rf"({json_obj_after}|{json_list_after})"

    try:
        comment = r"(//|\#)[^\n]*\n"
        comments = rf"({comment})+"
        s = re.sub(between_lines(comments), "\n", s)
        return json.dumps(json.loads(s), indent=4)
    except json.JSONDecodeError:
        ...

    try:
        # ... typically added by LLMs to identify that sequence may be continued
        s = re.sub(between_lines(r"\.\.\.\n"), "\n", s)
        return json.dumps(json.loads(s), indent=4)
    except json.JSONDecodeError:
        ...

    try:
        # missing comma between strings on separate lines
        s = re.sub(r"\"\s*\n\s*\"", '",\n"', s)
        return json.dumps(json.loads(s), indent=4)
    except json.JSONDecodeError:
        ...

    try:
        # Redundant trailing comma
        s = re.sub(
            r"((?<=[\"\d])|(?<=null)|(?<=true)|(?<=false))\s*\,(?=\s*[\}\]])", "", s
        )
        return json.dumps(json.loads(s), indent=4)
    except json.JSONDecodeError:
        ...

    try:
        # Fix incorrect quotes
        s = re.sub(rf"({json_before}\')|(\'{json_after})", '"', s)
        s = re.sub(rf"\'\s*\,\s*{json_after}", '", ', s)
        s = re.sub(r"\'\s*\:\s*(?=[\'\"])", '": ', s)
        s = re.sub(r"(?<=[\'\"])\s*\:\s*\'", ': "', s)
        return json.dumps(json.loads(s), indent=4)
    except json.JSONDecodeError:
        ...

    try:
        # Python-style values instead of JSON (inside fields)
        mapping = {"False": "false", "True": "true", "None": "null"}
        for pythonic, jsonic in mapping.items():
            s = re.sub(rf"\"\:\s*{pythonic}(?=\s*[\,\}}])", f'": {jsonic}', s)
        return json.dumps(json.loads(s), indent=4)
    except json.JSONDecodeError:
        ...

    try:
        # Drop inline comments
        s = re.sub(r"\/\*[^\n]*\*\/", "", s)
        return json.dumps(json.loads(s), indent=4)
    except json.JSONDecodeError:
        ...

    # incomplete JSON
    if s.startswith("{") and not s.endswith("}"):
        # count number of "
        if s.count('"') % 2 == 1:
            s += '"'
        s += "}"
    return s


def parse_json(
    input_string: str, raise_errors: bool = True, required_fields: list[str] = None
) -> list | dict | float | int | str:
    """
    Extract and parse JSON from AI-generated string.
    Args:
        input_string (str): String containing JSON.
        raise_errors (bool, optional):
            If True, raises exception on error, otherwise returns False on error.
        required_fields (list, optional): List of expected field names to validate the JSON object.
    Returns:
        list | dict | float | int | str: Parsed JSON data.
    """
    assert isinstance(required_fields, list) or required_fields is None
    try:
        if not input_string:  # empty string
            raise BadAIJsonAnswer(
                "Input string is empty. Cannot parse JSON from an empty string."
            )
        s = unwrap_json_substring(input_string)
        try:
            res = json.loads(s)
        except json.JSONDecodeError:
            res = json.loads(fix_json(s))
        if required_fields:
            if not isinstance(res, dict):
                raise BadAIJsonAnswer(
                    f"Expected a JSON object, but received: {type(res).__name__}"
                )
            for field in required_fields:
                if field not in res:
                    raise BadAIJsonAnswer(
                        f'Required field "{field}" is missing in the JSON object.'
                    )
        return res
    except (json.decoder.JSONDecodeError, BadAIJsonAnswer) as e:
        if raise_errors:
            raise BadAIJsonAnswer(str(e)) from e
        return False



import re

from ..configuration import Config


def make_remove_hidden_output(config: Config) -> callable:
    pattern = re.compile(
        f"{config.HIDDEN_OUTPUT_BEGIN}.*?{config.HIDDEN_OUTPUT_END}", flags=re.DOTALL
    )

    def remove_hidden_output(text: str) -> str:
        return pattern.sub("", text)

    return remove_hidden_output


def prepare_callbacks(config: Config, args, set_stream: bool = True) -> list[callable]:
    callbacks = args.pop("callbacks", []) or [] + config.CALLBACKS or []
    if "callback" in args:
        cb = args.pop("callback")
        if cb:
            callbacks.append(cb)
    if set_stream and "stream" not in args:
        args["stream"] = bool(callbacks)

    return callbacks



import asyncio
import builtins
import dataclasses
import inspect
import json
import os
import sys
import re
import subprocess
import logging
from dataclasses import dataclass
from fnmatch import fnmatch
from functools import lru_cache
from pathlib import Path
from typing import Any, Union, Callable

import tiktoken
from colorama import Fore

from .configuration import Config
from .types import BadAIAnswer, BadAIJsonAnswer
from .message_types import UserMsg, SysMsg, AssistantMsg
from .json_parsing import parse_json


def is_chat_model(model: str, config: Config = None) -> bool:
    """Detects if model is chat model or text completion model"""
    if config and config.CHAT_MODE is not None:
        return config.CHAT_MODE
    completion_keywords = ["instruct", "davinci", "babbage", "curie", "ada"]
    return not any(keyword in str(model).lower() for keyword in completion_keywords)


def is_image_model(model: str) -> bool:
    model = str(model)
    return model.startswith("dall-e-") or model.startswith("gpt-image-")


class ConvertableToMessage:
    @property
    def as_user(self) -> UserMsg:
        return UserMsg(str(self))

    @property
    def as_system(self) -> SysMsg:
        return SysMsg(str(self))

    @property
    def as_assistant(self) -> AssistantMsg:
        return AssistantMsg(str(self))

    @property
    def as_model(self) -> AssistantMsg:
        return self.as_assistant


class ExtendedString(str):
    """
    Provides a way of extending string with attributes and methods
    """

    def __new__(cls, string: str, attrs: dict = None):
        """
        Allows string to have attributes.
        """
        obj = str.__new__(cls, string)
        if attrs:
            for k, v in attrs.items():
                setattr(obj, k, v)
        return obj

    def __getattr__(self, item):
        """
        Provides chaining of global functions
        """
        global_func = inspect.currentframe().f_back.f_globals.get(item) or vars(
            builtins
        ).get(item, None)
        if callable(global_func):

            def method_handler(*args, **kwargs):
                res = global_func(self, *args, **kwargs)
                if isinstance(res, str) and not isinstance(res, ExtendedString):
                    res = ExtendedString(res)
                return res

            return method_handler

        # If there's not a global function with that name, raise an AttributeError as usual
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{item}'"
        )

    def to_tokens(
        self, for_model: str = None, encoding: str | tiktoken.Encoding = None
    ) -> list[int]:
        """
        Converts string to list of tokens.
        """
        from .tokenizing import encode

        return encode(self, for_model=for_model, encoding=encoding)

    def num_tokens(
        self, for_model: str = None, encoding: str | tiktoken.Encoding = None
    ) -> int:
        """
        Calculates quantity of tokens in target string.
        """
        return len(self.to_tokens(for_model=for_model, encoding=encoding))

    def parse_json(
        self, raise_errors: bool = True, required_fields: list[str] = None
    ) -> list | dict | float | int | str:
        return parse_json(self, raise_errors, required_fields)

    def contains_valid_json(self) -> bool:
        try:
            self.parse_json(raise_errors=True)
            return True
        except BadAIJsonAnswer:
            return False


class DataclassEncoder(json.JSONEncoder):
    """@private"""

    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return _default(self, o)


_default = json.JSONEncoder.default
json.JSONEncoder.default = DataclassEncoder().default


def parse(
    text: str, field_format: str = r"\[\[(.*?)\]\]", required_fields: list = None
) -> dict:
    """
    Parse a document divided into sections and convert it into a dictionary.
    """
    pattern = rf"{field_format}\n(.*?)(?=\n{field_format}|$)"
    matches = re.findall(pattern, text, re.DOTALL)
    result = {key.strip().lower(): value for key, value, _ in matches}
    if required_fields:
        for field in required_fields:
            if field not in result:
                raise BadAIAnswer(f"Field '{field}' is required but not found")
    return result


def file_link(file_path: str | Path):
    """Returns file name in format displayed in PyCharm console as a link."""
    return "file:///" + str(Path(file_path).absolute()).replace("\\", "/")


def list_files(
    target_dir: str | Path = "",
    exclude: list[str | Path] = None,
    relative_to: str | Path = None,
    absolute: bool = False,
    posix: bool = False,
) -> list[Path]:
    """
    Lists files in a specified directory, excluding those that match given patterns.

    This function traverses the specified directory recursively and returns a list of all files
    that do not match the specified exclusion patterns. It can return absolute paths,
    paths relative to the target directory, or paths relative to a specified directory.

    Args:
        target_dir (str | Path): The directory to search in.
        exclude (list[str | Path]): Patterns of files to exclude.
        relative_to (str | Path, optional): Base directory for relative paths.
            If None, paths are relative to `target_dir`. Defaults to None.
        absolute (bool, optional): If True, returns absolute paths. Defaults to False.
        posix (bool, optional): If True, returns posix paths. Defaults to False.

    Returns:
        list[Path]: A list of Path objects representing the files found.

    Example:
        exclude_patterns = ['*.pyc', '__pycache__/*']
        target_directory = '/path/to/target'
        files = list_files(target_directory, exclude_patterns)
    """
    def _is_file(file_path: Path) -> bool:
        if sys.version_info >= (3, 13):
            return file_path.is_file(follow_symlinks=False)
        try:
            return file_path.is_file()
        except OSError:
            return False

    exclude = exclude or []
    target = Path(target_dir or os.getcwd()).resolve()
    relative_to = Path(relative_to).resolve() if relative_to else None
    if absolute and relative_to is not None:
        raise ValueError(
            "list_files(): Cannot specify both 'absolute' and 'relative_to'. Choose one."
        )

    return [
        p.as_posix() if posix else p
        for p in (
            path.resolve() if absolute else path.relative_to(relative_to or target)
            for path in target.rglob("*")
            if _is_file(path)
            and not any(
                fnmatch(str(path.relative_to(target)), str(pattern))
                for pattern in exclude
            )
        )
    ]


def is_kaggle() -> bool:
    """
    Returns True if the code is running in a Kaggle notebook.
    """
    return "KAGGLE_KERNEL_RUN_TYPE" in os.environ


def is_notebook() -> bool:
    """
    Returns True if the code is running in a Jupyter notebook.
    """
    return "ipykernel" in sys.modules


def is_google_colab() -> bool:
    """
    Returns True if the code is running in a Google Colab notebook
    """
    return "google.colab" in sys.modules


def get_vram_usage(as_string=True, color=Fore.GREEN):
    """
    Returns GPU VRAM usage as a string or a list of objects.
    """

    @dataclass
    class _MemUsage:
        name: str
        used: int
        free: int
        total: int

    cmd = (
        "nvidia-smi"
        " --query-gpu=name,memory.used,memory.free,memory.total"
        " --format=csv,noheader,nounits"
    )
    try:
        out = subprocess.check_output(cmd, shell=True, text=True).strip()

        mu = [
            _MemUsage(*[i.strip() for i in line.split(",")])
            for line in out.splitlines()
        ]
        if not as_string:
            return mu
        c, r = (color, Fore.RESET) if color else ("", "")
        return "\n".join(
            [
                f"GPU: {c}{i.name}{r}, "
                f"VRAM: {c}{i.used}{r}/{c}{i.total}{r} MiB used, "
                f"{c}{i.free}{r} MiB free"
                for i in mu
            ]
        )
    except subprocess.CalledProcessError:
        msg = "No GPU found or nvidia-smi is not installed"
        return f"{Fore.RED}{msg}{Fore.RESET}" if as_string else None


def show_vram_usage():
    """Prints GPU VRAM usage."""
    print(get_vram_usage(as_string=True))


def return_default(default, *args):
    if isinstance(default, type) and issubclass(default, BaseException):
        raise default()
    if isinstance(default, BaseException):
        raise default
    if inspect.isbuiltin(default):
        return default(*args)
    if inspect.isfunction(default):
        arg_count = default.__code__.co_argcount
        if inspect.ismethod(default):
            arg_count -= 1
        return default(*args) if arg_count >= len(args) else default()

    return default


def extract_number(
    text: str,
    default=None,
    position="last",
    dtype: type | str = float,
    rounding: bool = False,
) -> int | float | Any:
    """
    Extract a number from a string.
    """
    assert position in ["last", "first"], f"Invalid position: {position}"
    idx = {"last": -1, "first": 0}[position]

    dtype = {"int": int, "float": float}.get(dtype, dtype)
    assert dtype in [int, float], f"Invalid dtype: {dtype}"
    if rounding:
        dtype = float
    regex = {int: r"[-+]?\d+", float: r"[-+]?\d*\.?\d+"}[dtype]

    numbers = re.findall(regex, str(text))
    if numbers:
        try:
            value = dtype(numbers[idx].strip())
            return round(value) if rounding else value
        except (ValueError, OverflowError):
            ...
    return return_default(default, text)


def dedent(text: str) -> str:
    """
    Removes minimal shared leading whitespace from each line
    and strips leading and trailing empty lines.
    """
    lines = text.splitlines()
    while lines and lines[0].strip() == "":
        lines.pop(0)
    while lines and lines[-1].strip() == "":
        lines.pop()
    non_empty_lines = [line for line in lines if line.strip()]
    if non_empty_lines:
        min_indent = min((len(line) - len(line.lstrip())) for line in non_empty_lines)
        dedented_lines = [
            line[min_indent:] if line and len(line) >= min_indent else line
            for line in lines
        ]
    else:
        dedented_lines = lines
    return "\n".join(dedented_lines)


RETURN_EXCEPTION = object()


async def run_parallel(
    tasks: list,
    max_concurrent_tasks: int,
    allow_failures: bool = False,
    return_on_failure: Any = RETURN_EXCEPTION,
    log_errors: bool = True,
):
    """
    Run tasks in parallel with a limit on the number of concurrent tasks.
    Args:
        tasks (list): A list of awaitable tasks to run.
        max_concurrent_tasks (int): Maximum number of tasks to run concurrently.
        allow_failures (bool): If True, exceptions in tasks are caught and
            handled according to the `return_on_failure` parameter.
        return_on_failure (Any): Value to return for failed tasks if `allow_failures` is True.
            If set to microcore.utils.RETURN_EXCEPTION, the exception object is returned.
        log_errors (bool): If True, errors are logged when `allow_failures` is True.
    """
    semaphore = asyncio.Semaphore(max_concurrent_tasks)

    async def worker(task):
        async with semaphore:
            try:
                return await task
            except Exception as e:
                if allow_failures:
                    if log_errors:
                        logging.error(e)
                    return e if return_on_failure == RETURN_EXCEPTION else return_on_failure
                raise

    return await asyncio.gather(*[worker(task) for task in tasks])


class CantResolveCallable(ValueError):
    """
    Raised when a callable cannot be resolved by name.
    """
    def __init__(self, message: str = None, name: str = None, e: Exception = None):
        message = message or f"Can't resolve callable by name '{name}'{', ' + str(e) if e else ''}"
        super().__init__(message)
        self.name = name


@lru_cache
def _load_callable(fn_path: str) -> callable:
    try:
        if "." not in fn_path:
            fn = globals()[fn_path]
        else:
            parts = fn_path.split(".")
            # Try resolve as *module.ClassName.static_method if 1st character is upper-cased
            if len(parts) >= 3 and len(parts[-2]) and parts[-2][0].isupper():
                module_name = ".".join(parts[:-2])
                class_name = parts[-2]
                try:
                    module = __import__(module_name, fromlist=[class_name])
                    cls = getattr(module, class_name)
                    fn = getattr(cls, parts[-1])
                    assert callable(fn)
                    return fn
                except (ImportError, AttributeError, AssertionError, ValueError):
                    pass
            module_name = ".".join(parts[:-1])
            func_name = parts[-1]
            if not module_name:
                raise CantResolveCallable(f"Invalid module name: {module_name}")
            module = __import__(module_name, fromlist=[func_name])
            fn = getattr(module, func_name)
        assert callable(fn)
    except (ImportError, AttributeError, AssertionError, ValueError) as e:
        raise CantResolveCallable(name=fn_path, e=e) from e
    return fn


def resolve_callable(
    fn: Union[Callable, str, None], allow_empty=False
) -> Union[Callable, None]:
    """
    Resolves a callable function from a string.
    Supported formats:
      - module[.submodules].function
      - module[.submodules].ClassName.static_method
    """
    if callable(fn):
        return fn
    if not fn:
        if allow_empty:
            return None
        raise CantResolveCallable("Can't resolve callable: function is not specified")
    return _load_callable(fn)


def levenshtein(a: str, b: str) -> int:
    """Compute the Levenshtein edit distance between two strings.

    The **Levenshtein distance** is the minimum number of single‑character
    edits (insertions, deletions, or substitutions) required to transform one
    string into the other.

    This implementation uses the classic Wagner–Fischer dynamic‑programming
    algorithm and stores only a single row of the DP matrix at any time,
    reducing memory usage to be linear in the length of the shorter string.

    Args:
        a (str): First input string.
        b (str): Second input string.

    Returns:
        int: Non‑negative integer representing the edit distance. A value of
        ``0`` means the strings are identical.

    Complexity:
        * **Time** ``O(ab)``
        * **Space** ``O(min(a, b))``

    Examples:
        >>> levenshtein("kitten", "sitting")
        3
        >>> levenshtein("graph", "giraffe")
        4
    """
    if a == b:
        return 0
    # Ensure a is the shorter string to reduce memory
    if len(a) > len(b):
        a, b = b, a
    previous = list(range(len(a) + 1))
    for i, ch_b in enumerate(b, start=1):
        current = [i]
        for j, ch_a in enumerate(a, start=1):
            cost = 0 if ch_a == ch_b else 1
            current.append(
                min(
                    current[-1] + 1,        # insertion
                    previous[j] + 1,        # deletion
                    previous[j - 1] + cost  # substitution
                )
            )
        previous = current
    return previous[-1]


def most_similar(
    needle: str,
    haystack: list[str],
    distance_func: callable = levenshtein,
    case_sensitive: bool = False,
) -> tuple[str, int]:
    """
    Find the most similar string from a list of strings using the
    specified distance function.

    Args:
        needle (str): The word to compare against.
        haystack (list[str]): A list of words to compare with.
        distance_func (callable): The distance function to use for comparison.
            Defaults to levenshtein.
        case_sensitive (bool): If True, the comparison is case-sensitive.

    Returns:
        tuple[str, int]: A tuple containing the most similar word and its distance
            from the given word.

    Raises:
        ValueError: If haystack is empty.
    """
    if not haystack:
        raise ValueError("Haystack cannot be empty")

    min_dist = float('inf')
    most_similar_word = None
    a = needle if case_sensitive else needle.lower()
    for word in haystack:
        b = word if case_sensitive else word.lower()
        dist = distance_func(a, b)
        if dist < min_dist:
            min_dist = dist
            most_similar_word = word

    return most_similar_word, min_dist



import json
import logging
import os
from dataclasses import dataclass, field, fields
from enum import Enum
from pathlib import Path
from typing import Any, Union, Callable

import dotenv
from colorama import Fore

_MISSING = object()

TRUE_VALUES = ["1", "TRUE", "YES", "ON", "ENABLED", "Y", "+"]
"""@private"""

PRINT_STREAM = "print_stream"
"""Logging method ID, value for 'config.use_logging'"""

DEFAULT_LOCAL_ENV_FILE = ".env"


def from_env(default=None, dtype=None):
    """
    Provides default value for the configuration dataclass
    from the environment variable with the name equal to field name in upper case"""
    return field(
        default=_MISSING, metadata=dict(_from_env=True, _default=default, _dtype=dtype)
    )


def get_bool_from_env(env_var: str, default: bool | None = False) -> bool | None:
    """Convert value of environment variable to boolean"""
    if env_var not in os.environ:
        return default
    return os.getenv(env_var, str(default)).upper() in TRUE_VALUES


def get_object_from_env(env_var: str, dtype: type, default: Any = None):
    val_from_env = os.getenv(env_var, _MISSING)  # pylint: disable=W1508
    if isinstance(val_from_env, str):
        val_from_env = val_from_env.strip()
        if val_from_env:
            try:
                val_from_env = json.loads(val_from_env.strip())
                assert isinstance(
                    val_from_env, dtype
                ), f"Expected {dtype.__name__}, got {type(val_from_env).__name__}"
            except (json.JSONDecodeError, AssertionError) as e:
                raise LLMConfigError(
                    f"Invalid value in environment variable '{env_var}'. "
                    f"Expected: JSON {dtype.__name__}, received: '{val_from_env}'"
                ) from e
        else:
            val_from_env = _MISSING
    if val_from_env is _MISSING:
        if default is None:  # instead of default factory
            default = dtype()
        val_from_env = default
    return val_from_env


class ApiType(str, Enum):
    """LLM API types"""

    OPEN_AI = "open_ai"
    AZURE = "azure"
    """See https://learn.microsoft.com/en-us/azure/ai-services/openai/concepts/models"""
    ANYSCALE = "anyscale"
    """See https://www.anyscale.com/endpoints"""
    DEEP_INFRA = "deep_infra"
    """List of text generation models: https://deepinfra.com/models?type=text-generation"""
    ANTHROPIC = "anthropic"
    GOOGLE_VERTEX_AI = "google_vertex_ai"
    GOOGLE_AI_STUDIO = "google_ai_studio"

    # Local models
    FUNCTION = "function"
    TRANSFORMERS = "transformers"
    NONE = "none"

    @staticmethod
    def is_local(api_type: str) -> bool:
        return api_type in (ApiType.FUNCTION, ApiType.TRANSFORMERS, ApiType.NONE)

    def __str__(self):
        return self.value


class EmbeddingDbType(str, Enum):
    CHROMA = "chroma"
    QDRANT = "qdrant"
    NONE = ""

    def __str__(self):
        return self.value


_default_dotenv_loaded = False


@dataclass
class BaseConfig:
    """Base class for configuration dataclasses"""

    USE_DOT_ENV: bool = None
    DOT_ENV_FILE: str | Path = None

    def __post_init__(self):
        self._dot_env_setup()
        self._update_from_env()

    def _dot_env_setup(self):
        global _default_dotenv_loaded

        if self.USE_DOT_ENV is None:
            self.USE_DOT_ENV = get_bool_from_env("USE_DOT_ENV", True)

        if self.USE_DOT_ENV:
            if self.DOT_ENV_FILE or not _default_dotenv_loaded:
                fp = self.DOT_ENV_FILE
                if fp and "~" in str(fp):
                    fp = Path(fp).expanduser()
                dotenv.load_dotenv(
                    override=True,
                    dotenv_path=fp
                )
            if not self.DOT_ENV_FILE:
                _default_dotenv_loaded = True

    def _update_from_env(self):
        for f in fields(self):
            if f.metadata.get("_from_env") and getattr(self, f.name) is _MISSING:
                env_name = f.name.upper()
                default = f.metadata["_default"]
                dtype = f.metadata.get("_dtype")
                if dtype is bool:
                    val_from_env = get_bool_from_env(env_name, default)
                elif dtype in [dict, list]:
                    val_from_env = get_object_from_env(env_name, dtype, default)
                else:
                    val_from_env = os.getenv(env_name, default)

                setattr(self, f.name, val_from_env)

    def __iter__(self):
        for f in fields(self):
            value = getattr(self, f.name)
            yield f.name, value


@dataclass
class _OpenAIEnvVars:
    # OS Environment variables expected by OpenAI library
    # Will be used as defaults for LLM
    # @todo: implement lib_defaults to take default values from openai lib if available
    OPENAI_API_TYPE: str = from_env(ApiType.OPEN_AI)
    OPENAI_API_KEY: str = from_env("")
    OPENAI_API_BASE: str = from_env("https://api.openai.com/v1")
    OPENAI_API_VERSION: str = from_env()


@dataclass
class _AnthropicEnvVars:
    ANTHROPIC_API_KEY: str = from_env()


@dataclass
class _GoogleVertexAiEnvVars:
    GOOGLE_VERTEX_ACCESS_TOKEN: str = from_env()
    GOOGLE_VERTEX_PROJECT_ID: str = from_env()
    GOOGLE_VERTEX_LOCATION: str = from_env()
    GOOGLE_VERTEX_GCLOUD_AUTH: bool = from_env(dtype=bool)

    GOOGLE_VERTEX_RESPONSE_VALIDATION: bool = from_env(dtype=bool, default=False)
    GOOGLE_GEMINI_SAFETY_SETTINGS: dict = from_env(dtype=dict)


@dataclass
class LLMConfig(BaseConfig, _OpenAIEnvVars, _AnthropicEnvVars, _GoogleVertexAiEnvVars):
    """LLM configuration"""

    LLM_API_TYPE: str = from_env()
    """
    See `ApiType`.
    To use services that is not listed in `ApiType`,
    but provides OpenAPI interface, use `ApiType.OPEN_AI`"""

    LLM_API_KEY: str = from_env("")
    LLM_API_BASE: str = from_env()
    """Base URL for the LLM API, e.g. https://api.openai.com/v1"""

    LLM_API_VERSION: str = from_env()
    LLM_DEPLOYMENT_ID: str = from_env()
    """Required by `ApiType.AZURE`"""

    MODEL: str = from_env()
    """Language model name"""

    TIKTOKEN_ENCODING: str = from_env()
    """Will enforce using specific encoding for token size measurement"""

    LLM_DEFAULT_ARGS: dict = from_env(dtype=dict)
    """
    You may specify here default arguments for the LLM API calls,
     i. e. temperature, max_tokens, etc.
     """

    AZURE_DEPLOYMENT_ID: str = from_env()

    INFERENCE_FUNC: Union[Callable, str] = from_env()
    """Inference function for local models"""
    CHAT_MODE: bool = from_env(dtype=bool)
    """Is it a chat or completion model"""
    INIT_PARAMS: dict = from_env(dtype=dict)
    """Custom initialization parameters for the model"""

    HIDDEN_OUTPUT_BEGIN: str = from_env()
    HIDDEN_OUTPUT_END: str = from_env()
    """Remove <think>...</think> from LLM response for models like DeepSeek R1"""
    CALLBACKS: list[Callable] = field(default_factory=list)

    VALIDATE_CONFIG: bool = from_env(dtype=bool, default=True)

    def __post_init__(self):
        super().__post_init__()
        self._init_llm_options()
        self.VALIDATE_CONFIG and self.validate()

    def uses_local_model(self) -> bool:
        return ApiType.is_local(self.LLM_API_TYPE)

    def hiding_output(self) -> bool:
        return bool(self.HIDDEN_OUTPUT_BEGIN and self.HIDDEN_OUTPUT_END)

    def _init_llm_options(self):
        if self.INFERENCE_FUNC:
            if not self.LLM_API_TYPE:
                self.LLM_API_TYPE = ApiType.FUNCTION
        if self.uses_local_model():
            return

        # Use defaults from ENV variables expected by OpenAI API
        self.LLM_API_TYPE = self.LLM_API_TYPE or self.OPENAI_API_TYPE

        if self.LLM_API_TYPE == ApiType.AZURE:
            self.LLM_API_VERSION = self.LLM_API_VERSION or self.OPENAI_API_VERSION
            self.LLM_DEPLOYMENT_ID = self.LLM_DEPLOYMENT_ID or self.AZURE_DEPLOYMENT_ID
        elif self.LLM_API_TYPE == ApiType.GOOGLE_AI_STUDIO:
            self.MODEL = self.MODEL or "gemini-pro"
        elif self.LLM_API_TYPE == ApiType.GOOGLE_VERTEX_AI:
            self.MODEL = self.MODEL or "gemini-1.0-pro"
            if self.GOOGLE_VERTEX_GCLOUD_AUTH is None:
                self.GOOGLE_VERTEX_GCLOUD_AUTH = get_bool_from_env(
                    "GOOGLE_VERTEX_GCLOUD_AUTH", not self.GOOGLE_VERTEX_ACCESS_TOKEN
                )
        elif self.LLM_API_TYPE == ApiType.ANYSCALE:
            self.LLM_API_BASE = (
                self.LLM_API_BASE or "https://api.endpoints.anyscale.com/v1"
            )
            self.MODEL = self.MODEL or "meta-llama/Llama-2-70b-chat-hf"
        elif self.LLM_API_TYPE == ApiType.DEEP_INFRA:
            self.LLM_API_BASE = (
                self.LLM_API_BASE or "https://api.deepinfra.com/v1/openai"
            )
            self.MODEL = self.MODEL or "meta-llama/Llama-2-70b-chat-hf"
        elif self.LLM_API_TYPE == ApiType.ANTHROPIC:
            self.LLM_API_BASE = self.LLM_API_BASE or "https://api.anthropic.com/"
            self.MODEL = self.MODEL or "claude-3-opus-20240229"
            self.LLM_API_KEY = self.LLM_API_KEY or self.ANTHROPIC_API_KEY
        else:
            self.LLM_API_BASE = self.LLM_API_BASE or self.OPENAI_API_BASE
            self.LLM_API_KEY = self.LLM_API_KEY or self.OPENAI_API_KEY
            self.LLM_API_VERSION = self.LLM_API_VERSION or self.OPENAI_API_VERSION
            self.MODEL = self.MODEL or "gpt-3.5-turbo"

    def _validate_local_llm(self):
        if self.CHAT_MODE is None:
            logging.warning(
                "When using local models, "
                "(bool)CHAT_MODE configuration option should be explicitly set"
            )
        if self.LLM_API_TYPE == ApiType.FUNCTION:
            if not self.INFERENCE_FUNC:
                raise LLMConfigError(
                    "INFERENCE_FUNC should be provided for local models"
                )
        elif self.LLM_API_TYPE == ApiType.TRANSFORMERS:
            if not self.MODEL:
                raise LLMConfigError(
                    "MODEL should be provided for local transformers models"
                )

    def validate(self):
        """
        Validate LLM configuration

        Raises:
            LLMConfigError
        """
        if self.LLM_API_TYPE == ApiType.NONE:
            return
        if self.uses_local_model():
            self._validate_local_llm()
            return
        if self.INFERENCE_FUNC:
            raise LLMConfigError(
                "INFERENCE_FUNC should be provided only for local models"
            )
        if self.LLM_API_TYPE == ApiType.GOOGLE_VERTEX_AI:
            if (
                not self.GOOGLE_VERTEX_ACCESS_TOKEN
                and not self.GOOGLE_VERTEX_GCLOUD_AUTH
            ):
                raise LLMConfigError(
                    "GOOGLE_VERTEX_ACCESS_TOKEN should be provided "
                    "or GOOGLE_VERTEX_GCLOUD_AUTH should be enabled"
                )
        else:
            if not self.LLM_API_KEY:
                raise LLMApiKeyError()
            if self.LLM_API_TYPE == ApiType.AZURE:
                if not self.LLM_API_BASE:
                    raise LLMApiBaseError()
                if not self.LLM_DEPLOYMENT_ID:
                    raise LLMApiDeploymentIdError()
                if not self.LLM_API_VERSION:
                    raise LLMApiVersionError()

    def describe(self, return_dict=False):
        """
        Informal description of the configuration
        """
        prev_env = os.environ.copy()
        os.environ.clear()
        default = Config(LLM_API_TYPE=ApiType.NONE, USE_DOT_ENV=False)
        os.environ.update(prev_env)
        data = {
            k.lower().replace("llm_", ""): v
            for k, v in dict(self).items()
            if v is not None and v != getattr(default, k) and k != "USE_DOT_ENV"
        }
        for k, v in data.items():
            if "_key" in k and isinstance(v, str):
                if len(v) <= 3:
                    continue
                data[k] = v[: 1 if len(v) <= 12 else 3] + "****" + v[-2:]
        if return_dict:
            return data

        print("Config:")
        for k, v in data.items():
            print(f"  {k}: {Fore.GREEN}{v}{Fore.RESET}")
        return None


class LLMConfigError(ValueError):
    """LLM configuration error"""
    BASE_MSG = "LLM configuration error"

    def __init__(self, message: str = None):
        message = f"{self.BASE_MSG}: {message}"
        super().__init__(message)


class LLMApiKeyError(LLMConfigError):
    """LLM API KEY error"""

    def __init__(self, message: str = None):
        message = message or "LLM_API_KEY is absent"
        super().__init__(message)


class LLMApiBaseError(LLMConfigError):
    """LLM API BASE error"""

    def __init__(self, message: str = None):
        message = message or "LLM_API_BASE is required for using Azure models"
        super().__init__(message)


class LLMApiDeploymentIdError(LLMConfigError):
    """LLM API DEPLOYMENT ID error"""

    def __init__(self, message: str = None):
        message = message or "LLM_DEPLOYMENT_ID is required for using Azure models"
        super().__init__(message)


class LLMApiVersionError(LLMConfigError):
    """LLM API VERSION error"""

    def __init__(self, message: str = None):
        message = message or "LLM_API_VERSION is required for using Azure models"
        super().__init__(message)


@dataclass
class Config(LLMConfig):
    """MicroCore configuration"""

    USE_LOGGING: bool = from_env(default=False)
    """Whether to use logging or not, see `microcore.use_logging`"""

    PROMPT_TEMPLATES_PATH: str | Path = from_env("tpl")
    """Path to the folder with prompt templates, ./tpl by default"""

    STORAGE_PATH: str | Path = from_env("storage")
    """Path to the folder with file storage, ./storage by default"""

    STORAGE_DEFAULT_FILE_EXT: str = from_env(default="")

    EMBEDDING_DB_FOLDER: str = from_env(default="embedding_db")
    """Folder within microcore.config.Config.STORAGE_PATH for storing embeddings"""

    EMBEDDING_DB_FUNCTION: Any = from_env()

    EMBEDDING_DB_ALLOW_DUPLICATES: bool = from_env(dtype=bool, default=False)

    EMBEDDING_DB_HOST: str = from_env(default=None)

    EMBEDDING_DB_PORT: str = from_env(default=None)

    EMBEDDING_DB_TYPE: str = from_env(EmbeddingDbType.CHROMA)

    EMBEDDING_DB_TIMEOUT: int = from_env(default=10 * 60)

    EMBEDDING_DB_SIZE: int = from_env(default=0)
    """Used with Qdrant"""

    DEFAULT_ENCODING: str = from_env("utf-8")
    """Used in file system operations, utf-8 by default"""

    JINJA2_AUTO_ESCAPE: bool = from_env(dtype=bool, default=False)

    ELEVENLABS_API_KEY: str = from_env()

    TEXT_TO_SPEECH_PATH: str | Path = from_env()
    """Path to the folder with generated voice files"""

    MAX_CONCURRENT_TASKS: int = from_env(default=None)

    SAVE_MEMORY: bool = from_env(dtype=bool, default=False)
    """
    Some additional data will not be collected:
      - LLMResponse objects will not contain the links to the prompt field
    """

    AI_SYNTAX_FUNCTION_NAME_FIELD: str = from_env(default="call")

    JINJA2_GLOBALS: dict = from_env(dtype=dict)

    MCP_SERVERS: list = from_env(dtype=list)

    INTERACTIVE_SETUP: bool = field(default=False)
    """Whether to run interactive setup if configuration is not valid."""

    def __post_init__(self):
        # Enforce using .env files if interactive_setup is enabled
        if self.INTERACTIVE_SETUP and not self.DOT_ENV_FILE:
            self.DOT_ENV_FILE = DEFAULT_LOCAL_ENV_FILE
        try:
            super().__post_init__()
        except LLMConfigError as e:
            if self.INTERACTIVE_SETUP and not os.path.exists(self.DOT_ENV_FILE):
                from .interactive_setup import interactive_setup
                orig_logging = self.USE_LOGGING  # avoid rewr. logging settings
                config = interactive_setup(self.DOT_ENV_FILE)
                self.__dict__.update(config.__dict__)
                self.USE_LOGGING = orig_logging
            else:
                raise e
        if self.TEXT_TO_SPEECH_PATH is None:
            self.TEXT_TO_SPEECH_PATH = Path(self.STORAGE_PATH) / "voicing"

    def validate(self):
        super().validate()
        if self.EMBEDDING_DB_TYPE == EmbeddingDbType.QDRANT:
            if not self.EMBEDDING_DB_SIZE:
                raise LLMConfigError(
                    "EMBEDDING_DB_SIZE is required configuration parameter for Qdrant"
                )
            if not self.EMBEDDING_DB_HOST:
                raise LLMConfigError(
                    "EMBEDDING_DB_HOST is required configuration parameter for Qdrant"
                )
            if not self.EMBEDDING_DB_FUNCTION:
                raise LLMConfigError(
                    "EMBEDDING_DB_FUNCTION is required configuration parameter for Qdrant"
                )



import asyncio
from google.ai.generativelanguage import Content, Part
from google.generativeai import GenerationConfig
from google.generativeai.types import HarmCategory, HarmBlockThreshold
import google.generativeai as genai
from ..configuration import Config
from .._prepare_llm_args import prepare_chat_messages
from ..message_types import Role
from ..types import LLMAsyncFunctionType, LLMFunctionType, BadAIAnswer
from ..wrappers.llm_response_wrapper import LLMResponse
from .shared import prepare_callbacks


async def _a_process_streamed_response(response, callbacks: list[callable]):
    response_text: str = ""
    async for chunk in response:
        if text_chunk := chunk.text:
            response_text += text_chunk
            for cb in callbacks:
                if asyncio.iscoroutinefunction(cb):
                    await cb(text_chunk)
                else:
                    cb(text_chunk)
    return LLMResponse(response_text, {})


def _process_streamed_response(response, callbacks: list[callable]):
    response_text: str = ""
    for chunk in response:
        if text_chunk := chunk.text:
            response_text += text_chunk
            [cb(text_chunk) for cb in callbacks]
    return LLMResponse(response_text, {})


def make_llm_functions(config: Config) -> tuple[LLMFunctionType, LLMAsyncFunctionType]:
    genai.configure(api_key=config.LLM_API_KEY, **config.INIT_PARAMS)
    if config.GOOGLE_GEMINI_SAFETY_SETTINGS is None:
        # Only new categories
        config.GOOGLE_GEMINI_SAFETY_SETTINGS = {
            HarmCategory.HARM_CATEGORY_HARASSMENT: HarmBlockThreshold.BLOCK_NONE,
            HarmCategory.HARM_CATEGORY_HATE_SPEECH: HarmBlockThreshold.BLOCK_NONE,
            HarmCategory.HARM_CATEGORY_SEXUALLY_EXPLICIT: HarmBlockThreshold.BLOCK_NONE,
            HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT: HarmBlockThreshold.BLOCK_NONE,
        }

    def _prepare_chat(prompt, **kwargs):
        model_name = kwargs.pop("model", config.MODEL)
        callbacks = prepare_callbacks(config, kwargs, set_stream=False)
        kwargs.pop("stream", None)
        model = genai.GenerativeModel(
            model_name,
            generation_config=GenerationConfig(**kwargs),
            safety_settings=config.GOOGLE_GEMINI_SAFETY_SETTINGS,
        )
        messages = _chat_messages_to_google(prepare_chat_messages(prompt))
        last_message = messages.pop()
        chat = model.start_chat(history=messages)
        return chat, last_message.parts[0], callbacks

    async def allm(prompt, **kwargs):
        chat, msg, callbacks = _prepare_chat(prompt, **kwargs)
        try:
            response = await chat.send_message_async(msg, stream=bool(callbacks))
            if callbacks:
                return await _a_process_streamed_response(response, callbacks)
            return LLMResponse(response.text, response.__dict__)
        except ValueError as e:
            raise BadAIAnswer(str(e)) from e

    def llm(prompt, **kwargs):
        chat, msg, callbacks = _prepare_chat(prompt, **kwargs)
        try:
            response = chat.send_message(msg, stream=bool(callbacks))
            if callbacks:
                return _process_streamed_response(response, callbacks)
            return LLMResponse(response.text, response.__dict__)
        except ValueError as e:
            raise BadAIAnswer(str(e)) from e

    return llm, allm


def _chat_messages_to_google(messages: list[dict]):
    # Convert roles to Google Vertex roles
    # (system,user,assistant) -> (user, model)
    messages = [msg.copy() for msg in messages]
    for msg in messages:
        if msg["role"] == Role.SYSTEM:
            msg["role"] = "user"
        elif msg["role"] == Role.ASSISTANT:
            msg["role"] = "model"

    # Merge sequences of messages with same role
    # to avoid following error:
    #   google.api_core.exceptions.InvalidArgument:
    #   400 Please ensure that multiturn requests ends with a user role or a function response.
    merged_msg = []
    for msg in messages:
        if merged_msg and msg["role"] == merged_msg[-1]["role"]:
            merged_msg[-1]["content"] += "\n" + msg["content"]
        else:
            merged_msg.append(msg)
    vertex_messages = [
        Content(role=msg["role"], parts=[Part(text=msg["content"])])
        for msg in merged_msg
    ]
    return vertex_messages



from dataclasses import asdict

from .message_types import DEFAULT_MESSAGE_ROLE, Msg
from .types import TPrompt


def prepare_prompt(prompt) -> str:
    """Converts prompt to string for LLM completion API"""
    return "\n".join(
        [
            str(p["content"]) if isinstance(p, dict) and "content" in p else str(p)
            for p in (prompt if isinstance(prompt, list) else [prompt])
        ]
    )


def prepare_chat_messages(prompt: TPrompt) -> list[dict]:
    """Converts prompt to messages for LLM chat API (OpenAI)"""
    messages: list = prompt if isinstance(prompt, list) else [prompt]
    return [
        (
            dict(role=DEFAULT_MESSAGE_ROLE, content=msg)
            if isinstance(msg, str)
            else (
                asdict(msg, dict_factory=msg.DICT_FACTORY)
                if isinstance(msg, Msg)
                else msg
            )
        )
        for msg in messages
    ]



"""Configuration presets collection."""

from .configuration import PRINT_STREAM, DEFAULT_LOCAL_ENV_FILE, ApiType, EmbeddingDbType

# Minimal handy configuration preset
# for non-production usage (simple scripts, small experiments, etc.)
MIN_SETUP = dict(
    INTERACTIVE_SETUP=True,
    DOT_ENV_FILE=DEFAULT_LOCAL_ENV_FILE,
    USE_LOGGING=PRINT_STREAM,
)

# No LLM, No embedding DB
EMPTY = dict(
    LLM_API_TYPE=ApiType.NONE,
    EMBEDDING_DB_TYPE=EmbeddingDbType.NONE
)



import asyncio
import logging
import datetime
from typing import Optional, TYPE_CHECKING
from dataclasses import dataclass, field
from enum import Enum

import httpx
import requests
import mcp.types

from .utils import ExtendedString, ConvertableToMessage
from .ai_func import AiFuncSyntax
from . import ui
from .types import BadAIAnswer, BadAIJsonAnswer
from .wrappers.llm_response_wrapper import LLMResponse
from ._env import env
from .file_storage import storage
from .json_parsing import parse_json

# Temporary mitigate dotenv warnings
# because of issue in FastMCP: https://github.com/jlowin/fastmcp/issues/2018
logging.getLogger("dotenv.main").setLevel(logging.CRITICAL)
if TYPE_CHECKING:
    from fastmcp.client import Client
    from fastmcp.client.progress import ProgressHandler


class WrongMcpUsage(BadAIAnswer):
    ...


class HeaderAuth(httpx.Auth):
    def __init__(self, header_name_or_dict: str | dict, header_value: str | None = None):
        if isinstance(header_name_or_dict, dict):
            if header_value is not None:
                raise ValueError(
                    "HeaderAuth: If a dictionary is provided in first argument, "
                    "header_value should not be set."
                )
            self.headers = header_name_or_dict
        else:
            if header_value is None:
                raise ValueError(
                    "HeaderAuth: header_value must be provided when header_name is a string"
                )
            self.headers = {str(header_name_or_dict): str(header_value)}

    def auth_flow(self, request):
        for k, v in self.headers.items():
            request.headers[k] = v
        yield request


class ToolsCache:
    FILE = "mcp/tools.json"

    @staticmethod
    def read(mcp_url: str) -> Optional["Tools"]:
        logging.info(f"Checking MCP tools cache for {ui.green(mcp_url)}...")
        tools = storage.read_json(ToolsCache.FILE, default={}).get(mcp_url, None)
        if tools is not None:
            return Tools.from_list([Tool(**raw) for raw in tools.values()])
        return None

    @staticmethod
    def write(mcp_url: str, tools: "Tools"):
        logging.info(f"Storing MCP tools cache for {ui.green(mcp_url)}...")
        cached_tools = storage.read_json(ToolsCache.FILE, default={})
        cached_tools[mcp_url] = tools
        storage.write_json(ToolsCache.FILE, cached_tools)

    @staticmethod
    def clear():
        logging.info("Clearing MCP tools cache...")
        storage.delete(ToolsCache.FILE)


class MCPAnswer(ExtendedString, ConvertableToMessage):
    ...


class McpTransport(str, Enum):
    SSE: str = "sse"
    STREAMABLE_HTTP: str = "streamable_http"
    WS: str = "ws"
    STDIO: str = "stdio"

    def __str__(self):
        return self.value


@dataclass
class MCPConnection:
    url: str = None
    transport: McpTransport = field(default=McpTransport.STREAMABLE_HTTP)
    tools: Optional["Tools"] = field(default=None, init=False)
    _client: Optional["Client"] = field(default=None, init=False)

    @staticmethod
    async def init(
        url: str,
        transport: McpTransport,
        fetch_tools: bool = True,
        use_cache: bool = True,
        connect_timeout: float = 10,
        auth: httpx.Auth = None,
    ) -> "MCPConnection":
        # Temporary mitigate FastMCP errors in some environments
        # https://github.com/jlowin/fastmcp/issues/2018
        from fastmcp.client import Client
        con: MCPConnection = MCPConnection(url=url, transport=transport)
        con._client = Client(  # pylint: disable=W0212
            url,
            timeout=connect_timeout,
            auth=auth,
        )
        await con._client.__aenter__()  # pylint: disable=E1101,W0212,C2801
        if fetch_tools:
            await con.fetch_tools(use_cache=use_cache)
        return con

    async def close(self):
        if self._client:
            await self._client.__aexit__(None, None, None)
            self._client = None
        else:
            logging.error(f"Trying to close MCP connection that is not opened ({self.url})")

    async def fetch_tools(self, use_cache: bool = True) -> "Tools":
        if self.tools is not None:
            return self.tools

        if use_cache and (cached_tools := ToolsCache.read(self.url)):
            logging.info("Using MCP tools from cache for %s", ui.green(self.url))
            self.tools = cached_tools
            return self.tools

        logging.info("Fetching tools from MCP %s", ui.green(self.url))
        mcp_tools = await self._client.list_tools()
        self.tools = Tools.from_list([Tool.from_mcp(tool) for tool in mcp_tools])
        if use_cache:
            self.update_tools_cache()
        return self.tools

    def update_tools_cache(self):
        if self.tools is None:
            raise RuntimeError("Tools are not fetched yet. Call fetch_tools() first.")
        ToolsCache.write(self.url, self.tools)

    async def call(
        self,
        name: str,
        timeout: datetime.timedelta | float | int | None = None,
        progress_handler: Optional["ProgressHandler"] = None,
        **kwargs
    ):
        assert env().config.AI_SYNTAX_FUNCTION_NAME_FIELD not in kwargs
        params = dict(kwargs)
        params[env().config.AI_SYNTAX_FUNCTION_NAME_FIELD] = name
        return await self.exec(params, timeout=timeout, progress_handler=progress_handler)

    async def exec(
        self,
        params: dict | LLMResponse,
        timeout: datetime.timedelta | float | int | None = None,
        progress_handler: Optional["ProgressHandler"] = None,
    ):
        tool_field = env().config.AI_SYNTAX_FUNCTION_NAME_FIELD
        if isinstance(params, LLMResponse):
            try:
                params = params.parse_json(
                    raise_errors=True,
                    required_fields=[tool_field],
                )
            except BadAIJsonAnswer as e:
                raise WrongMcpUsage(str(e)) from e
        elif isinstance(params, str):
            try:
                params = parse_json(
                    params,
                    raise_errors=True,
                    required_fields=[tool_field]
                )
            except BadAIJsonAnswer as e:
                raise WrongMcpUsage(str(e)) from e
        params = dict(params)
        if not (name := params.pop(tool_field, "")):
            raise WrongMcpUsage(
                f"Wrong MCP tool usage:\n"
                f"Tool name should be passed in '{tool_field}' field."
            )
        logging.info(f"Calling MCP tool {ui.green(name)} with {params}...")
        call_tool_result: mcp.types.CallToolResult = await self._client.call_tool(
            name=name,
            arguments=params,
            timeout=timeout,
            progress_handler=progress_handler
        )
        if (
            call_tool_result
            and len(call_tool_result.content) == 1
            and call_tool_result.content[0].type == "text"
        ):
            return MCPAnswer(call_tool_result.content[0].text, dict(response=call_tool_result))
        return call_tool_result

    # alias
    execute = exec


@dataclass
class Tool:
    @dataclass
    class Arg:
        name: str = field()
        description: str = field(default="")
        type: str = field(default="string")
        required: bool = field(default=...)
        default: any = field(default=...)

        def __post_init__(self):
            if self.required is ...:
                self.required = self.default is ...
            if self.default is ...:
                self.default = None

    name: str = field()
    description: str = field(default="")
    args: dict[str, Arg | dict] = field(default_factory=dict)

    def __post_init__(self):
        for key in self.args.keys():  # pylint: disable=C0201, C0206
            if isinstance(self.args[key], dict):
                self.args[key] = Tool.Arg(**self.args[key])

    @staticmethod
    def from_mcp(tool: mcp.types.Tool) -> "Tool":
        t = Tool(
            name=tool.name,
            description=tool.description,
        )
        for param_name, data in tool.inputSchema.get("properties", {}).items():
            param = Tool.Arg(
                name=param_name,
                description=data.get("title", ""),
                type=data.get("type", "string"),
                required=param_name in tool.inputSchema.get("required", []),
                default="",
            )
            t.args[param_name] = param
        return t

    def describe(self, syntax: AiFuncSyntax = None) -> str:
        syntax = syntax or AiFuncSyntax.DEFAULT
        tpl_file = f"ai-func.{syntax}.j2" if syntax in AiFuncSyntax else syntax
        metadata = self._get_metadata()
        return env().tpl_function(tpl_file, **metadata)

    def _get_metadata(self):
        return dict(
            name=self.name,
            description=self.description,
            args={
                arg.name: dict(
                    default="NOT_SET" if arg.required else arg.default,
                    type=arg.type,
                    comment=arg.description,  # @todo rename it to description
                )
                for arg in self.args.values()
            }
        )

    def __str__(self):
        return self.describe(syntax=AiFuncSyntax.DEFAULT)


class Tools(dict[str, Tool]):
    def __str__(self):
        return "\n".join([str(tool) for tool in self.values()])

    @staticmethod
    def from_list(tools: list[Tool]) -> "Tools":
        return Tools({tool.name: tool for tool in tools})


@dataclass
class MCPServer:
    url: str
    name: str = field(default="")
    tools: Tools = field(default_factory=Tools)
    transport: McpTransport = field(default=None)
    auth: httpx.Auth | dict | None = field(default=None)

    @staticmethod
    def _try_sse_or_streamable_http(url) -> tuple[McpTransport, str]:
        if url.endswith("/"):
            test_sse_url = f"{url}sse"
        else:
            test_sse_url = f"{url}/sse"
        try:
            response = requests.request(method="HEAD", url=test_sse_url, timeout=5)
            if response.status_code == 200:
                return McpTransport.SSE, test_sse_url
        except requests.RequestException:
            # not a SSE endpoint or not reachable
            pass
        return McpTransport.STREAMABLE_HTTP, f"{url}/mcp"

    @staticmethod
    def _guess_transport_type_by_url(url: str) -> Optional[McpTransport]:
        if url.startswith("ws://") or url.startswith("wss://"):
            return McpTransport.WS
        if url.endswith("/mcp") or url.endswith("/mcp/"):
            return McpTransport.STREAMABLE_HTTP
        if url.endswith("/sse") or url.endswith("/sse/"):
            return McpTransport.SSE
        return None

    @staticmethod
    def name_from_url(url: str) -> str:
        """Domain name from URL."""
        return url.split("//")[-1].split("/")[0]

    def __post_init__(self):
        if not self.name:
            self.name = MCPServer.name_from_url(self.url)
        if not self.transport:
            self.transport = self._guess_transport_type_by_url(self.url)
        if self.auth and isinstance(self.auth, dict):
            self.auth = HeaderAuth(self.auth)

    async def connect(
        self,
        fetch_tools: bool = True,
        use_cache: bool = True,
        connect_timeout: float = 10,
    ) -> MCPConnection:
        if self.transport:
            transport, url = self.transport, self.url
        else:
            transport, url = self._try_sse_or_streamable_http(self.url)
        return await MCPConnection.init(
            url=url,
            transport=transport,
            fetch_tools=fetch_tools,
            use_cache=use_cache,
            connect_timeout=connect_timeout,
            auth=self.auth
        )

    def get_tools_cache(self) -> Tools | None:
        return ToolsCache.read(self.url)


class MCPRegistry(dict[str, MCPServer]):
    def __init__(self, server_configs: list[dict | str]):
        super().__init__()
        for server_config in server_configs:
            if isinstance(server_config, str):
                server_config = {
                    "name": MCPServer.name_from_url(server_config),
                    "url": server_config
                }
            self[server_config["name"]] = MCPServer(**server_config)

    def get(self, server_name: str) -> MCPServer:
        if server_name not in self:
            raise ValueError(f"MCP server '{server_name}' not found in registry")
        return self[server_name]

    async def precache_tools(
        self,
        raise_errors: bool = False,
        connect_timeout: int = 10,
    ):
        async def precache_server_tools(server_name):
            conn = None
            try:
                conn = await self.get(server_name).connect(
                    fetch_tools=True,
                    use_cache=False,
                    connect_timeout=connect_timeout,
                )
                conn.update_tools_cache()
            except Exception as e:  # pylint: disable=W0718
                logging.error("Failed to precache tools for MCP server %s: %s", server_name, e)
                if raise_errors:
                    raise
            finally:
                if conn is not None:
                    await conn.close()

        await asyncio.gather(*[precache_server_tools(srv) for srv in self.keys()])

    async def connect_to(
        self,
        server_name: str,
        fetch_tools: bool = True,
        use_cache: bool = True,
    ) -> MCPConnection:
        mcp_server = self.get(server_name)
        return await mcp_server.connect(fetch_tools=fetch_tools, use_cache=use_cache)


def server(name: str) -> MCPServer:
    """
    Returns MCP server by name from the registry.

    Args:
        name (str): The name of the MCP server.

    Returns:
        MCPServer: The MCP server instance.

    Raises:
        ValueError: If the server with the given name is not found in the registry.
    """
    return env().mcp_registry.get(name)



import os.path
from dataclasses import dataclass, field, asdict, fields
from importlib.util import find_spec
from typing import TYPE_CHECKING

import jinja2

from .embedding_db import AbstractEmbeddingDB
from .configuration import (
    Config,
    ApiType,
    LLMConfigError,
    EmbeddingDbType,
    PRINT_STREAM,
)
from .presets import MIN_SETUP
from .types import TplFunctionType, LLMAsyncFunctionType, LLMFunctionType
from .templating.jinja2 import make_jinja2_env, make_tpl_function
from .llm.openai_llm import make_llm_functions as make_openai_llm_functions
from .llm.local_llm import make_llm_functions as make_local_llm_functions

if TYPE_CHECKING:
    from .wrappers.llm_response_wrapper import LLMResponse  # noqa: F401
    from transformers import PreTrainedModel, PreTrainedTokenizer  # noqa: F401
    from .mcp import MCPRegistry


@dataclass
class Env:
    config: Config = field(default_factory=Config)
    jinja_env: jinja2.Environment = None
    tpl_function: TplFunctionType = None
    llm_async_function: LLMAsyncFunctionType = None
    llm_function: LLMFunctionType = None
    llm_before_handlers: list[callable] = field(default_factory=list)
    llm_after_handlers: list[callable] = field(default_factory=list)
    texts: AbstractEmbeddingDB = None
    model: "PreTrainedModel" = field(
        default=None, init=False, repr=False
    )  # noqa
    tokenizer: "PreTrainedTokenizer" = field(  # noqa
        default=None, init=False, repr=False
    )
    _mcp_registry: "MCPRegistry" = field(init=False, default=None)

    def __post_init__(self):
        global _env
        _env = self
        self.init_templating()
        self.init_llm()
        if self.config.USE_LOGGING:
            from .logging import use_logging

            use_logging(stream=self.config.USE_LOGGING == PRINT_STREAM)
        self.init_similarity_search()

    def make_stopping_criteria(self, seq: str | list[str]) -> list[callable]:
        raise NotImplementedError

    def init_templating(self):
        self.jinja_env = make_jinja2_env(self)
        self.tpl_function = make_tpl_function(self)

    @property
    def mcp_registry(self):
        if self._mcp_registry is None:
            from .mcp import MCPRegistry
            self._mcp_registry = MCPRegistry(self.config.MCP_SERVERS)
        return self._mcp_registry

    def init_llm(self):
        if self.config.LLM_API_TYPE == ApiType.NONE:

            def not_configured(*args, **kwargs) -> "LLMResponse":
                raise LLMConfigError("Language model is not configured")

            async def a_not_configured(*args, **kwargs) -> "LLMResponse":
                raise LLMConfigError("Language model is not configured")

            self.llm_function, self.llm_async_function = (
                not_configured,
                a_not_configured,
            )

        elif self.config.LLM_API_TYPE == ApiType.FUNCTION:
            self.llm_function, self.llm_async_function = make_local_llm_functions(
                self.config
            )
        elif self.config.LLM_API_TYPE == ApiType.TRANSFORMERS:
            try:
                from .llm.local_transformers import (
                    make_llm_functions as make_transformers_llm_functions,
                )
            except ModuleNotFoundError as e:
                raise ModuleNotFoundError(
                    "To use local Transformers language models, "
                    "you need to install the `transformers`, `pytorch` and `accelerate` packages. "
                ) from e
            (
                self.llm_function,
                self.llm_async_function,
            ) = make_transformers_llm_functions(self.config, self)
        elif self.config.LLM_API_TYPE == ApiType.ANTHROPIC:
            try:
                from .llm.anthropic import (
                    make_llm_functions as make_anthropic_llm_functions,
                )
            except ModuleNotFoundError as e:
                raise ModuleNotFoundError(
                    "To use the Anthropic language models, "
                    "you need to install the `anthropic` package. "
                    "Run `pip install anthropic`."
                ) from e
            self.llm_function, self.llm_async_function = make_anthropic_llm_functions(
                self.config
            )
        elif self.config.LLM_API_TYPE == ApiType.GOOGLE_VERTEX_AI:
            try:
                from .llm.google_vertex_ai import (
                    make_llm_functions as make_google_vertex_llm_functions,
                )
            except ModuleNotFoundError as e:
                raise ModuleNotFoundError(
                    "To use the Google Vertex language models, "
                    "you need to install the `vertexai` package "
                    "and authenticate with Google Cloud cli."
                    "Run `pip install vertexai`."
                ) from e
            (
                self.llm_function,
                self.llm_async_function,
            ) = make_google_vertex_llm_functions(self.config)
        elif self.config.LLM_API_TYPE == ApiType.GOOGLE_AI_STUDIO:
            try:
                from .llm.google_genai import (
                    make_llm_functions as make_google_genai_llm_functions,
                )
            except ModuleNotFoundError as e:
                raise ModuleNotFoundError(
                    "To use the Google Gemini language models via AI Studio, "
                    "you need to install the `google-generativeai` package. "
                    "Run `pip install google-generativeai`."
                ) from e
            (
                self.llm_function,
                self.llm_async_function,
            ) = make_google_genai_llm_functions(self.config)
        else:
            self.llm_function, self.llm_async_function = make_openai_llm_functions(
                self.config
            )

    def init_similarity_search(self):
        if (
            self.config.EMBEDDING_DB_TYPE == EmbeddingDbType.CHROMA
            and find_spec("chromadb") is not None
        ):
            from .embedding_db.chromadb import ChromaEmbeddingDB

            self.texts = ChromaEmbeddingDB(self.config)
            return

        if self.config.EMBEDDING_DB_TYPE == EmbeddingDbType.QDRANT:
            if find_spec("qdrant_client") is None:
                raise ModuleNotFoundError(
                    "To use Qdrant, install the `qdrant-client` package. "
                    "Run `pip install qdrant-client`."
                )
            from .embedding_db.qdrant import QdrantEmbeddingDB
            self.texts = QdrantEmbeddingDB(self.config)
            return


@dataclass
class _Configure(Config):
    def __post_init__(self):
        global _env
        _env = None
        super().__post_init__()
        Env(self)


configure: callable = _Configure
"""Applies configuration to MicroCore environment"""

if True:  # pylint: disable=W0125
    # This block is inside a condition to avoid breaking IDE autocompletion

    _fields = list(map(lambda f: f.name, fields(Config)))

    def _config_builder_wrapper(cfg: Config | dict | str = None, **kwargs):
        """
        - Convert configuration keys to uppercase
        - Add LLM_ prefix to keys if necessary
        - Allow to configure from Config instance or dictionary
        """
        if cfg:
            assert not kwargs, "Cannot pass both cfg and kwargs"
        if isinstance(cfg, dict):
            return _config_builder_wrapper(**cfg)
        if isinstance(cfg, str):
            if not os.path.isfile(cfg):
                raise LLMConfigError(f"Configuration file not found: {cfg}")
            return _config_builder_wrapper(Config(USE_DOT_ENV=True, DOT_ENV_FILE=cfg))
        kwargs = {str(k).upper(): v for k, v in kwargs.items()}
        for k in list(kwargs.keys()):
            if not hasattr(Config, k) and (
                hasattr(Config, key := f"LLM_{k}") or key in _fields
            ):
                kwargs[key] = kwargs.pop(k)
        return _Configure(**(cfg and asdict(cfg) or kwargs))

    configure = _config_builder_wrapper


def min_setup():
    """
    Minimal handy setup for non-production usage
    (simple scripts, small experiments, etc.)
    """
    return configure(MIN_SETUP)


_env: Env | None = None


def env() -> Env:
    """Returns the current MicroCore environment"""
    return _env or Env()


def config() -> Config:
    """Resolve current configuration"""
    return env().config



import re
import logging
from datetime import datetime
from typing import Any

from .utils import run_parallel, RETURN_EXCEPTION
from .wrappers.llm_response_wrapper import LLMResponse, DictFromLLMResponse
from .types import TPrompt, LLMContextLengthExceededError
from ._env import env


def convert_exception(e: Exception, model: str = None) -> Exception | None:
    """
    Convert LLM exceptions microcore-specific exceptions if possible.
    Args:
        e (Exception): Original exception
        model (str): LLM model name, used for better error messages
    Returns:
        Converted exception or None if no conversion is possible
    """
    if not isinstance(e, Exception):
        return None
    t, msg = f"{type(e).__module__}.{type(e).__name__}", str(e)
    max_tokens, actual_tokens = None, None
    if t == "openai.BadRequestError" and "context_length_exceeded" in msg:
        match = re.search(
            r"maximum context length is (\d+) tokens.*?resulted in (\d+) tokens",
            msg
        )
        if match:
            max_tokens = int(match.group(1))
            actual_tokens = int(match.group(2))
        return LLMContextLengthExceededError(
            actual_tokens=actual_tokens,
            max_tokens=max_tokens,
            model=model
        )
    if t == "anthropic.BadRequestError" and "prompt is too long:" in msg:
        if match := re.search(r"(\d+)\s+tokens\s+>\s+(\d+)\s+maximum", msg):
            max_tokens = int(match.group(2))
            actual_tokens = int(match.group(1))
        return LLMContextLengthExceededError(
            actual_tokens=actual_tokens,
            max_tokens=max_tokens,
            model=model
        )
    if (
        t == "google.api_core.exceptions.InvalidArgument"
        and "The input token count exceeds the maximum number of tokens allowed" in msg
    ):
        if match := re.search(
            r"The input token count exceeds the maximum number of tokens allowed (\d+)",
            msg
        ):
            max_tokens = int(match.group(1))
        return LLMContextLengthExceededError(
            actual_tokens=actual_tokens,
            max_tokens=max_tokens,
            model=model
        )
    return None


def llm(
    prompt: TPrompt,
    retries: int = 0,
    parse_json: bool | dict = False,
    **kwargs
) -> str | LLMResponse:
    """
    Request Large Language Model synchronously

    Args:
        prompt (str | Msg | dict | list[str | Msg | dict]): Text to send to LLM
        retries (int): Number of retries in case of error
        parse_json (bool|dict):
            If True, parses response as JSON,
            alternatively non-empty dict can be used as parse_json arguments
        **kwargs (dict): Parameters supported by the LLM API

            See parameters supported by the OpenAI:

            - https://platform.openai.com/docs/api-reference/completions/create
            - https://platform.openai.com/docs/api-reference/chat/create

            **Additional parameters:**

                - callback: callable - callback function
                to be called on each chunk of text,
                enables response streaming if supported by the LLM API
                - callbacks: list[callable] - collection of callbacks
                to be called on each chunk of text,
                enables response streaming if supported by the LLM API

    Returns:

        Text generated by the LLM as string
        with all fields returned by API accessible as an attributes.

        See fields returned by the OpenAI:

        - https://platform.openai.com/docs/api-reference/completions/object
        - https://platform.openai.com/docs/api-reference/chat/object
    """
    [h(prompt, **kwargs) for h in env().llm_before_handlers]
    start = datetime.now()
    tries = retries + 1
    while tries > 0:
        try:
            tries -= 1
            response = env().llm_function(prompt, **kwargs)
            break
        except Exception as e:  # pylint: disable=W0718
            converted_exception = convert_exception(e)
            # If context length exceeded, or no tries left --> do not retry
            if tries == 0 or isinstance(converted_exception, LLMContextLengthExceededError):
                if converted_exception:
                    raise converted_exception from e
                raise e
            logging.error(f"LLM error: {e}")
            logging.info(f"Retrying... {tries} retries left")
            continue
    try:
        response.gen_duration = (datetime.now() - start).total_seconds()
        if not env().config.SAVE_MEMORY:
            response.prompt = prompt
    except AttributeError:
        ...
    [h(response) for h in env().llm_after_handlers]
    if tries > 0:
        retry_params = dict(**kwargs)
        retry_params["retries"] = tries - 1
        setattr(
            response,
            "_retry_callback",
            lambda: llm(prompt, **retry_params)
        )
    if parse_json:
        parsing_params = parse_json if isinstance(parse_json, dict) else {}
        return response.parse_json(**parsing_params)
    return response


async def allm(
    prompt: TPrompt,
    retries: int = 0,
    parse_json: bool | dict = False,
    **kwargs
) -> str | LLMResponse | DictFromLLMResponse:
    """
    Request Large Language Model asynchronously

    Args:
        prompt (str | Msg | dict | list[str | Msg | dict]): Text to send to LLM
        retries (int): Number of retries in case of error
        parse_json (bool|dict):
            If True, parses response as JSON,
            alternatively non-empty dict can be used as parse_json arguments
        **kwargs (dict): Parameters supported by the LLM API

            See parameters supported by the OpenAI:

            - https://platform.openai.com/docs/api-reference/completions/create
            - https://platform.openai.com/docs/api-reference/chat/create

            **Additional parameters:**

            - callback: callable - callback function
            to be called on each chunk of text,
            enables response streaming if supported by the LLM API
            - callbacks: list[callable] - collection of callbacks
            to be called on each chunk of text,
            enables response streaming if supported by the LLM API

            Note: async callbacks are supported only for async LLM API calls

    Returns:

        Text generated by the LLM as string
        with all fields returned by API accessible as an attributes.

        See fields returned by the OpenAI:

        - https://platform.openai.com/docs/api-reference/completions/object
        - https://platform.openai.com/docs/api-reference/chat/object
    """
    [h(prompt, **kwargs) for h in env().llm_before_handlers]
    start = datetime.now()
    tries = retries + 1
    while tries > 0:
        try:
            tries -= 1
            response = await env().llm_async_function(prompt, **kwargs)
            break
        except Exception as e:  # pylint: disable=W0718
            converted_exception = convert_exception(e)
            # If context length exceeded, or no tries left --> do not retry
            if tries == 0 or isinstance(converted_exception, LLMContextLengthExceededError):
                if converted_exception:
                    raise converted_exception from e
                raise e
            logging.error(f"LLM error: {e}")
            logging.info(f"Retrying... {tries} retries left")
            continue
    try:
        response.gen_duration = (datetime.now() - start).total_seconds()
        if not env().config.SAVE_MEMORY:
            response.prompt = prompt
    except AttributeError:
        ...
    [h(response) for h in env().llm_after_handlers]
    if parse_json:
        try:
            parsing_params = parse_json if isinstance(parse_json, dict) else {}
            return response.parse_json(**parsing_params)
        except Exception as e:  # pylint: disable=W0718
            if tries > 0:
                logging.error(f"LLM error: {e}")
                logging.info(f"Retrying... {tries} retries left")
                return await allm(prompt, retries=tries - 1, parse_json=parse_json, **kwargs)
    return response


async def llm_parallel(
    prompts: list[TPrompt],
    max_concurrent_tasks: int = None,
    allow_failures: bool = False,
    return_on_failure: Any = RETURN_EXCEPTION,
    log_errors: bool = True,
    **kwargs
) -> list[str | LLMResponse]:
    """
    Execute multiple LLM requests in parallel

    Returns (list[LLMResponse | str]): a list of responses in the same order as the prompts
    """
    tasks = [allm(prompt, **kwargs) for prompt in prompts]

    if max_concurrent_tasks is None:
        max_concurrent_tasks = int(env().config.MAX_CONCURRENT_TASKS or 0)
    if not max_concurrent_tasks:
        max_concurrent_tasks = len(tasks)

    return await run_parallel(
        tasks,
        max_concurrent_tasks=max_concurrent_tasks,
        allow_failures=allow_failures,
        return_on_failure=return_on_failure,
        log_errors=log_errors,
    )



import hashlib
import logging
import sys
from dataclasses import dataclass
import uuid

from qdrant_client import QdrantClient
from qdrant_client.http.models import (
    PointIdsList,
    CollectionInfo,
    Record,
    FieldCondition,
    Filter,
    MatchValue,
    MatchText
)
from qdrant_client.models import VectorParams, Distance, PointStruct, ScoredPoint

from ..configuration import Config
from .. import SearchResult, SearchResults, AbstractEmbeddingDB


def is_sentence_transformer(fn):
    return fn.__class__.__name__ == 'SentenceTransformer'


def prepare_embedding_function(fn):
    if is_sentence_transformer(fn):
        return lambda x: fn.encode(x).tolist()
    return fn


@dataclass
class QdrantEmbeddingDB(AbstractEmbeddingDB):
    config: Config
    embedding_function: callable = None
    client: QdrantClient = None

    def __post_init__(self):

        logging.info(
            "Connecting to Qdrant at %s:%s",
            self.config.EMBEDDING_DB_HOST,
            self.config.EMBEDDING_DB_PORT
        )
        self.client = QdrantClient(
            host=self.config.EMBEDDING_DB_HOST,
            port=self.config.EMBEDDING_DB_PORT or 6333,
            timeout=self.config.EMBEDDING_DB_TIMEOUT,
        )
        self.embedding_function = prepare_embedding_function(self.config.EMBEDDING_DB_FUNCTION)

    @classmethod
    def _wrap_results(cls, points: list[ScoredPoint | Record]) -> list[str | SearchResult]:
        return SearchResults(
            [
                SearchResult(
                    i.payload["_text"],
                    dict(
                        id=i.id,
                        distance=i.score if hasattr(i, "score") else 0,
                        metadata={k: v for k, v in i.payload.items() if k != "_text"},
                    ),

                )
                for i in points
            ]
        )

    @classmethod
    def _convert_where(  # pylint: disable=too-many-branches
        cls,
        where: dict | None,
        kwargs=None
    ) -> Filter | None:
        where_doc = kwargs and kwargs.get("where_document", {}).get("$contains", None)
        if isinstance(where, Filter):
            if where_doc:
                raise ValueError(
                    "Cannot use `where_document` with Filter object passed as `where` argument. "
                    "Please use a dictionary instead."
                )
            return where

        conditions = []
        _and = True
        if where:
            if "$or" in where:
                _and = False
                for i in where["$or"]:
                    for k, v in i.items():
                        conditions.append(FieldCondition(key=k, match=MatchValue(value=v)))
            elif "$and" in where:
                _and = True
                for i in where["$and"]:
                    for k, v in i.items():
                        conditions.append(FieldCondition(key=k, match=MatchValue(value=v)))
            else:
                for k, v in where.items():
                    conditions.append(FieldCondition(key=k, match=MatchValue(value=v)))

        # ChromaDB format
        if where_doc:
            if not _and:
                raise ValueError(
                    "Cannot use `where_document` with `$or` condition. "
                )
            conditions.append(
                FieldCondition(
                    key="_text",
                    match=MatchText(text=kwargs["where_document"]["$contains"])
                )
            )

        if not conditions:
            return None
        if _and:
            return Filter(must=conditions)
        return Filter(should=conditions)

    def search(
        self,
        collection: str,
        query: str | list,
        n_results: int = 5,
        where: dict = None,
        **kwargs,
    ) -> list[str | SearchResult]:
        if not self.collection_exists(collection):
            return SearchResults([])

        if not isinstance(query, str):
            raise ValueError("`query` must be a string")

        query_vector = self.embedding_function(query)
        where = self._convert_where(where, kwargs)
        kwargs.pop("where_document", None)
        hits = self.client.query_points(
            collection_name=collection,
            query=query_vector,
            limit=n_results,
            query_filter=where,
            **kwargs,
        )
        return self._wrap_results(hits.points)

    def save_many(self, collection: str, items: list[tuple[str, dict] | str]):
        if not self.collection_exists(collection):
            self._create_collection(collection)
        point_structs = []
        ids = set()
        unique = not self.config.EMBEDDING_DB_ALLOW_DUPLICATES
        for i in items:
            if isinstance(i, str):
                text = i
                metadata = dict()
            else:
                text = i[0]
                metadata = i[1] or {}
            metadata["_text"] = text
            if unique:
                new_id = str(uuid.UUID(hashlib.md5(text.encode()).hexdigest()))
                if new_id in ids:
                    continue
                ids.add(new_id)
            else:
                new_id = str(uuid.uuid4())
            point_structs.append(
                PointStruct(
                    id=new_id,
                    vector=self.embedding_function(text),
                    payload=metadata
                )
            )

        operation_info = self.client.upsert(
            collection_name=collection,
            wait=True,
            points=point_structs,
        )
        return operation_info

    def clear(self, collection: str):
        if self.collection_exists(collection):
            self.client.delete_collection(collection_name=collection)

    def count(self, collection: str) -> int:
        if self.collection_exists(collection):
            return self._get_collection(collection).points_count
        return 0

    def delete(self, collection: str, what: str | list[str] | dict):
        # pylint: disable=R0801, duplicate-code
        if not self.collection_exists(collection):
            return

        if isinstance(what, str):
            ids, where = [what], None
        elif isinstance(what, list):
            ids, where = what, None
        elif isinstance(what, dict):
            ids, where = None, what
        else:
            raise ValueError("Invalid `what` argument")
        if ids is not None:
            points_selector = PointIdsList(points=ids)
        else:
            points_selector = self._convert_where(where)
        self.client.delete(collection_name=collection, points_selector=points_selector)

    def get(
        self,
        collection: str,
        ids: list[str] | str = None,
        limit: int = None,
        offset: int = None,
        where: dict = None,
        **kwargs,
    ) -> list[str | SearchResult] | str | SearchResult | None:
        if not self.collection_exists(collection):
            return SearchResults([]) if not isinstance(ids, str) else None
        if ids:
            raise NotImplementedError("Getting by ids is not supported for Qdrant")
        where = self._convert_where(where, kwargs)
        kwargs.pop("where_document", None)
        search_results = self._wrap_results(
            self.client.scroll(
                collection,
                limit=limit or sys.maxsize - 1,
                offset=offset or 0,
                scroll_filter=where,
                **kwargs
            )[0]
        )
        if isinstance(ids, str):
            return search_results[0] if search_results else None
        return search_results

    def get_all(self, collection: str) -> list[str | SearchResult]:
        if not self.collection_exists(collection):
            return SearchResults([])
        return self._wrap_results(self.client.scroll(collection, limit=sys.maxsize - 1)[0])

    def collection_exists(self, collection: str) -> bool:
        return self.client.collection_exists(collection)

    def _create_collection(self, name: str):
        assert self.config.EMBEDDING_DB_SIZE > 0
        size = self.config.EMBEDDING_DB_SIZE
        distance = Distance.COSINE
        self.client.create_collection(
            collection_name=name,
            vectors_config=VectorParams(size=size, distance=distance),
        )

    def _get_collection(self, name: str, create: bool = False) -> CollectionInfo | None:
        if not self.collection_exists(name):
            if create:
                self._create_collection(name)
            else:
                return None
        return self.client.get_collection(name)



import dataclasses

from colorama import Fore, init

from .configuration import ApiType
from .utils import is_chat_model, is_notebook
from ._env import env, config
from ._prepare_llm_args import prepare_chat_messages, prepare_prompt


def _format_request_log_str(prompt, **kwargs) -> str:
    nl = "\n" if LoggingConfig.DENSE else "\n" + LoggingConfig.INDENT
    model = _resolve_model(**kwargs)
    out = (
        f"{LoggingConfig.COLOR_RESET}Requesting LLM "
        f"{Fore.MAGENTA}{model}{LoggingConfig.COLOR_RESET}:"
        + (" " if LoggingConfig.DENSE else "\n")
    )
    if is_chat_model(model, env().config):
        for msg in prepare_chat_messages(prompt):
            role, content = (
                (msg["role"], msg["content"])
                if isinstance(msg, dict)
                else dataclasses.astuple(msg)
            )
            nl2 = "\n" if LoggingConfig.DENSE else nl + LoggingConfig.INDENT
            content = (" " if LoggingConfig.DENSE else nl2) + nl2.join(
                content.split("\n")
            )
            out += (
                f"{'' if LoggingConfig.DENSE else LoggingConfig.INDENT}"
                f"[{role.capitalize()}]:"
                f"{LoggingConfig.PROMPT_COLOR}{content}{LoggingConfig.COLOR_RESET}\n"
            )
    else:
        lines = prepare_prompt(prompt).split("\n")
        out = (
            LoggingConfig.PROMPT_COLOR
            + (" " if LoggingConfig.DENSE else LoggingConfig.INDENT)
            + nl.join(lines)
            + LoggingConfig.COLOR_RESET
        )
        if out.endswith("\n"):
            out = out[:-1]
    if LoggingConfig.STRIP_REQUEST_LINES:
        start_lines, end_lines = LoggingConfig.STRIP_REQUEST_LINES
        max_lines = start_lines + end_lines
        lines = out.split("\n")
        if len(lines) > max_lines:
            out = "\n".join(
                lines[:start_lines]
                + [
                    f"{LoggingConfig.INDENT}{Fore.YELLOW}"
                    f"...(output was truncated)..."
                    f"{LoggingConfig.PROMPT_COLOR}"
                ]
                + (lines[-end_lines:] if end_lines else [])
            )
    return out


def _resolve_model(**kwargs):
    cfg = config()
    model = kwargs.get("model") or cfg.LLM_DEFAULT_ARGS.get("model") or cfg.MODEL
    if cfg.LLM_API_TYPE == ApiType.AZURE:
        model = f"azure:{model}"
    return model


def _format_response_log_str(out) -> str:
    nl = "\n" if LoggingConfig.DENSE else "\n" + LoggingConfig.INDENT
    out_indented = (" " if LoggingConfig.DENSE else nl) + nl.join(
        (out or "").split("\n")
    )
    return (
        f"{LoggingConfig.COLOR_RESET}LLM Response:"
        f"{LoggingConfig.RESPONSE_COLOR}{out_indented}{LoggingConfig.COLOR_RESET}"
    )


class LoggingConfig:
    PROMPT_COLOR = Fore.LIGHTGREEN_EX
    RESPONSE_COLOR = Fore.CYAN
    COLOR_RESET = Fore.RESET
    INDENT: str = " " * 4
    DENSE: bool = False
    OUTPUT_METHOD: callable = print
    REQUEST_FORMATTER: callable = _format_request_log_str
    RESPONSE_FORMATTER: callable = _format_response_log_str
    STRIP_REQUEST_LINES: tuple[int, int] | None = [40, 15]


def _log_request(prompt, **kwargs):
    LoggingConfig.OUTPUT_METHOD(_format_request_log_str(prompt, **kwargs))


def _log_response(out):
    LoggingConfig.OUTPUT_METHOD(_format_response_log_str(out))


_is_new_request = False


def _stream_log_request(prompt, **kwargs):
    global _is_new_request
    _is_new_request = True
    _log_request(prompt, **kwargs)


def _stream_log_response(out):
    global _is_new_request
    if _is_new_request:
        LoggingConfig.OUTPUT_METHOD(_format_response_log_str(out))
        _is_new_request = False
    else:
        out = f"{LoggingConfig.RESPONSE_COLOR}{out}{LoggingConfig.COLOR_RESET}"
        if not LoggingConfig.DENSE and LoggingConfig.INDENT:
            out = out.replace("\n", "\n" + LoggingConfig.INDENT)
        LoggingConfig.OUTPUT_METHOD(out)


def _print_no_nln(s):
    print(s, end='', flush=True)


def use_logging(stream: bool = False):
    """Turns on logging of LLM requests and responses to console."""
    if not is_notebook():
        init(strip=False)
    if stream:
        LoggingConfig.OUTPUT_METHOD = _print_no_nln
        # Remove non-streamable handlers if any
        if _log_request in env().llm_before_handlers:
            env().llm_before_handlers.remove(_log_request)
        if _log_response in env().llm_after_handlers:
            env().llm_after_handlers.remove(_log_response)
        # Add streamable handlers if not already present
        if _stream_log_request not in env().llm_before_handlers:
            env().llm_before_handlers.append(_stream_log_request)
        if _stream_log_response not in config().CALLBACKS:
            env().config.CALLBACKS.append(_stream_log_response)
    else:
        if _log_request not in env().llm_before_handlers:
            env().llm_before_handlers.append(_log_request)
        if _log_response not in env().llm_after_handlers:
            env().llm_after_handlers.append(_log_response)
        # Cleanup prev. streamable setup if any
        if LoggingConfig.OUTPUT_METHOD is _print_no_nln:
            LoggingConfig.OUTPUT_METHOD = print
        if _stream_log_request in env().llm_before_handlers:
            env().llm_before_handlers.remove(_stream_log_request)
        if _stream_log_response in config().CALLBACKS:
            config().CALLBACKS.remove(_stream_log_response)
